import os
import pathlib

import dask
import pytest
import yaml

import xorq as xo
import xorq.vendor.ibis as ibis
from xorq.caching import ParquetStorage
from xorq.common.utils.defer_utils import deferred_read_parquet
from xorq.ibis_yaml.compiler import ArtifactStore, BuildManager
from xorq.ibis_yaml.config import config
from xorq.ibis_yaml.sql import find_relations
from xorq.vendor.ibis.common.collections import FrozenOrderedDict


@pytest.mark.snapshot_check
def test_build_manager_expr_hash(t, build_dir, snapshot):
    build_manager = ArtifactStore(build_dir)
    actual = build_manager.get_expr_hash(t)
    snapshot.assert_match(actual, "build_manager_expr_hash.txt")


def test_build_manager_roundtrip(t, build_dir):
    build_manager = ArtifactStore(build_dir)
    yaml_dict = {"a": "string"}
    expr_hash = "dummy-value"
    build_manager.save_yaml(yaml_dict, expr_hash, "expr.yaml")

    with open(build_dir / expr_hash / "expr.yaml") as f:
        out = f.read()
    assert out == "a: string\n"
    result = build_manager.load_yaml(expr_hash, "expr.yaml")
    assert result == yaml_dict


def test_build_manager_paths(t, build_dir):
    new_path = build_dir / "new_path"

    assert not os.path.exists(new_path)
    build_manager = ArtifactStore(new_path)
    assert os.path.exists(new_path)

    build_manager.get_build_path("hash")
    assert os.path.exists(new_path / "hash")


def test_clean_frozen_dict_yaml(build_dir):
    build_manager = ArtifactStore(build_dir)
    data = FrozenOrderedDict(
        {"string": "text", "integer": 42, "float": 3.14, "boolean": True, "none": None}
    )

    expected_yaml = """string: text
integer: 42
float: 3.14
boolean: true
none: null
"""
    out_path = build_manager.save_yaml(data, "hash", "expr.yaml")
    result = out_path.read_text()

    assert expected_yaml == result


@pytest.mark.xfail(reason="MemTable is not serializable")
def test_ibis_compiler(t, build_dir):
    t = xo.memtable({"a": [0, 1], "b": [0, 1]})
    expr = t.filter(t.a == 1).drop("b")
    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)

    roundtrip_expr = compiler.load_expr(expr_hash)

    assert expr.execute().equals(roundtrip_expr.execute())


def test_ibis_compiler_parquet_reader(build_dir):
    backend = xo.duckdb.connect()
    parquet_path = xo.config.options.pins.get_path("awards_players")
    awards_players = deferred_read_parquet(
        backend, parquet_path, table_name="award_players"
    )
    expr = awards_players.filter(awards_players.lgID == "NL").drop("yearID", "lgID")
    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)
    roundtrip_expr = compiler.load_expr(expr_hash)

    assert expr.execute().equals(roundtrip_expr.execute())


def test_compiler_sql(build_dir, parquet_dir):
    backend = xo.datafusion.connect()
    awards_players = deferred_read_parquet(
        backend,
        str(parquet_dir / "awards_players.parquet"),
        table_name="awards_players",
    )
    expr = awards_players.filter(awards_players.lgID == "NL").drop("yearID", "lgID")

    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)
    _roundtrip_expr = compiler.load_expr(expr_hash)
    expected_relation = find_relations(awards_players)[0]
    expted_sql_hash = dask.base.tokenize(str(ibis.to_sql(expr)))[: config.hash_length]

    assert os.path.exists(build_dir / expr_hash / "sql.yaml")
    assert os.path.exists(build_dir / expr_hash / "metadata.json")
    metadata = compiler.artifact_store.read_json(build_dir, expr_hash, "metadata.json")

    assert "current_library_version" in metadata
    sql_text = pathlib.Path(build_dir / expr_hash / "sql.yaml").read_text()
    expected_result = (
        "queries:\n"
        "  main:\n"
        "    engine: datafusion\n"
        f"    profile_name: {expr._find_backend()._profile.hash_name}\n"
        "    relations:\n"
        f"    - {expected_relation}\n"
        "    options: {}\n"
        f"    sql_file: {expted_sql_hash}.sql\n"
    )
    assert sql_text == expected_result


def test_deferred_reads_yaml(build_dir):
    backend = xo.datafusion.connect()
    # Factor out the config path
    config_path = xo.config.options.pins.get_path("awards_players")
    awards_players = deferred_read_parquet(
        backend,
        config_path,
        table_name="awards_players",
    )
    expr = awards_players.filter(awards_players.lgID == "NL").drop("yearID", "lgID")

    # Get the dynamic relation and profile hash
    expected_relation = find_relations(awards_players)[0]
    expected_profile = backend._profile.hash_name

    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)
    _roundtrip_expr = compiler.load_expr(expr_hash)

    yaml_path = build_dir / expr_hash / "deferred_reads.yaml"
    assert os.path.exists(yaml_path)
    sql_text = pathlib.Path(yaml_path).read_text()

    sql_str = str(ibis.to_sql(awards_players))
    expected_sql_file = dask.base.tokenize(sql_str)[: config.hash_length] + ".sql"

    expected_read_path = str(config_path)

    expected_result = (
        "reads:\n"
        f"  {expected_relation}:\n"
        "    engine: datafusion\n"
        f"    profile_name: {expected_profile}\n"
        "    relations:\n"
        f"    - {expected_relation}\n"
        "    options:\n"
        "      method_name: read_parquet\n"
        "      name: awards_players\n"
        "      read_kwargs:\n"
        f"      - path: {expected_read_path}\n"
        "      - table_name: awards_players\n"
        f"    sql_file: {expected_sql_file}\n"
    )

    assert sql_text == expected_result


def test_ibis_compiler_expr_schema_ref(t, build_dir):
    t = xo.memtable({"a": [0, 1], "b": [0, 1]})
    expr = t.filter(t.a == 1).drop("b")
    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)

    with open(build_dir / expr_hash / "expr.yaml") as f:
        yaml_dict = yaml.safe_load(f)

    assert yaml_dict["expression"]["schema_ref"]


def test_multi_engine_deferred_reads(build_dir):
    con0 = xo.connect()
    con1 = xo.connect()
    con2 = xo.duckdb.connect()
    con3 = xo.connect()

    awards_players = xo.examples.awards_players.fetch(con0).into_backend(con1)
    batting = xo.examples.batting.fetch(con2).into_backend(con1)
    expr = (
        awards_players.join(batting, predicates=["playerID", "yearID", "lgID"])
        .into_backend(con3)
        .filter(xo._.G == 1)
    )
    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)

    roundtrip_expr = compiler.load_expr(expr_hash)

    assert expr.execute().equals(roundtrip_expr.execute())


def test_multi_engine_with_caching(build_dir):
    con0 = xo.connect()
    con1 = xo.connect()
    con2 = xo.duckdb.connect()
    con3 = xo.connect()

    awards_players = xo.examples.awards_players.fetch(con0).into_backend(con1).cache()
    batting = xo.examples.batting.fetch(con2).into_backend(con1).cache()
    expr = (
        awards_players.join(batting, predicates=["playerID", "yearID", "lgID"])
        .into_backend(con3)
        .filter(xo._.G == 1)
    )
    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)

    roundtrip_expr = compiler.load_expr(expr_hash)

    assert expr.execute().equals(roundtrip_expr.execute())


def test_multi_engine_with_caching_with_parquet(build_dir, tmp_path):
    con0 = xo.connect()
    con1 = xo.connect()
    con2 = xo.duckdb.connect()
    con3 = xo.connect()

    storage = ParquetStorage(source=con1, path=tmp_path)

    awards_players = (
        xo.examples.awards_players.fetch(con0).into_backend(con1).cache(storage=storage)
    )
    batting = xo.examples.batting.fetch(con1).into_backend(con2).cache(storage=storage)
    expr = (
        awards_players.join(batting, predicates=["playerID", "yearID", "lgID"])
        .into_backend(con3)
        .filter(xo._.G == 1)
    )
    compiler = BuildManager(build_dir)
    expr_hash = compiler.compile_expr(expr)

    roundtrip_expr = compiler.load_expr(expr_hash)

    assert expr.execute().equals(roundtrip_expr.execute())



import pandas as pd

import xorq as xo
from xorq import memtable


N = 100000
# if single float deferred partitions of train and test will be returned
# With proportions (1-test_size, test_size)
test_size = 0.25
# init table
table = memtable([(i, "val") for i in range(N)], columns=["key1", "val"])


train_table, test_table = xo.train_test_splits(
    table, unique_key="key1", test_sizes=test_size, num_buckets=N, random_seed=42
)

train_count = xo.execute(train_table.count())
test_count = xo.execute(test_table.count())
total = train_count + test_count
print(f"train ratio: {round(train_count / total, 2)}")
print(f"test ratio: {round(test_count / total, 2)}\n")


# If test sizes is a list of floats , mutually exclusive partitions will be returned
partition_info = {
    "hold_out": 0.1,
    "test": 0.2,
    "validation": 0.3,
    "training": 0.4,
}

partitions = tuple(
    xo.train_test_splits(
        table,
        unique_key="key1",
        test_sizes=list(partition_info.values()),
        num_buckets=N,
        random_seed=42,
    )
)
counts = pd.Series(xo.execute(p.count()) for p in partitions)
total = sum(counts)

for i, partition_name in enumerate(partition_info.keys()):
    print(f"{partition_name.upper()} Ratio: {round(counts[i] / total, 2)}")

name = "split"
c = xo.calc_split_column(
    table,
    unique_key="key1",
    test_sizes=list(partition_info.values()),
    num_buckets=N,
    random_seed=42,
    name=name,
)
other_counts = xo.execute(c.value_counts().order_by(c.get_name())).set_index(name)[
    f"{name}_count"
]
assert counts.equals(other_counts)



import functools
import operator
import os
from urllib.parse import unquote_plus

import pandas as pd
import toolz
from openai import OpenAI
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)

import xorq as xo
from xorq.flight.utils import (
    schema_concat,
    schema_contains,
)


@functools.cache
def get_client():
    client = OpenAI(
        api_key=os.environ["OPENAI_API_KEY"],
    )
    return client


request_timeout = 3


def extract_sentiment(text):
    @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(3))
    def completion_with_backoff(**kwargs):
        return get_client().chat.completions.create(**kwargs)

    if text == "":
        return "NEUTRAL"
    messages = [
        {
            "role": "system",
            "content": "You are an AI language model trained to analyze and detect the sentiment of hackernews forum comments.",
        },
        {
            "role": "user",
            "content": f"Analyze the following hackernews comment and determine if the sentiment is: positive, negative or neutral. "
            f"Return only a single word, either POSITIVE, NEGATIVE or NEUTRAL: {text}",
        },
    ]
    try:
        response = completion_with_backoff(
            model="gpt-3.5-turbo",
            messages=messages,
            max_tokens=30,
            temperature=0,
            timeout=request_timeout,
        )
        return response.choices[0].message.content
    except Exception as e:
        return f"ERROR: {e}"


@toolz.curry
def get_hackernews_sentiment_batch(df: pd.DataFrame, input_col, append_col):
    return df.assign(
        **{
            append_col: df[input_col].map(
                toolz.compose(extract_sentiment, unquote_plus)
            )
        }
    )


input_col = "text"
append_col = "sentiment"
schema_requirement = xo.schema({input_col: "!str"})
schema_append = xo.schema({append_col: "!str"})
maybe_schema_in = toolz.compose(schema_contains(schema_requirement), xo.schema)
maybe_schema_out = toolz.compose(
    operator.methodcaller("to_pyarrow"),
    schema_concat(to_concat=schema_append),
    xo.Schema.from_pyarrow,
)


do_hackernews_sentiment_udxf = xo.expr.relations.flight_udxf(
    process_df=get_hackernews_sentiment_batch(
        input_col=input_col, append_col=append_col
    ),
    maybe_schema_in=maybe_schema_in,
    maybe_schema_out=maybe_schema_out,
    name="HackerNewsSentimentAnalyzer",
)

hn = xo.examples.hn_posts_nano.fetch(table_name="hackernews")

expr = (
    hn.order_by(hn.time.desc())
    .filter(
        xo.or_(
            hn.text.cast(str).like("%ClickHouse%"),
            hn.title.cast(str).like("%ClickHouse%"),
        )
    )
    .select(hn.text)
    .limit(2)
    .pipe(do_hackernews_sentiment_udxf)
)

df = expr.execute()
print(df)



import operator

import toolz

import xorq as xo
import xorq.expr.relations as rel
import xorq.vendor.ibis.expr.operations as ops
from xorq.caching import SourceStorage
from xorq.common.utils.graph_utils import (
    find_all_sources,
    walk_nodes,
)


node_types = (
    ops.DatabaseTable,
    ops.SQLQueryResult,
    rel.CachedNode,
    rel.Read,
    rel.RemoteTable,
    # ExprScalarUDF has an expr we need to get to
    # FlightOperator has a dynamically generated connection: it should be passed a Profile instead
)


def make_expr():
    cons = (con0, con1, con2, con3) = (
        xo.connect(),
        xo.connect(),
        xo.duckdb.connect(),
        xo.connect(),
    )

    read_node0 = xo.examples.awards_players.fetch(con0)
    remote_node0 = read_node0.into_backend(con1)
    cached_node0 = remote_node0.cache(SourceStorage(source=con1))
    read_node1 = xo.examples.batting.fetch(con2)
    remote_node1 = read_node1.into_backend(con1)
    remote_node2 = cached_node0.join(
        remote_node1, predicates=["playerID", "yearID", "lgID"]
    ).into_backend(con3)
    cached_node1 = remote_node2.cache()
    expr = cached_node2 = cached_node1[lambda t: t.G == 1].cache()
    nodes = {
        rel.CachedNode: (
            cached_node0.op(),
            cached_node1.op(),
            cached_node2.op(),
        ),
        rel.Read: (
            read_node0.op(),
            read_node1.op(),
        ),
        rel.RemoteTable: (
            remote_node0.op(),
            remote_node1.op(),
            remote_node2.op(),
        ),
    }
    return (cons, nodes, expr)


def test_walk_nodes():
    (_, nodes, expr) = make_expr()
    node_types = tuple(nodes)
    walked_nodes = walk_nodes(node_types, expr)
    expected = sorted(
        ((k, set(v)) for k, v in nodes.items()),
        key=toolz.compose(operator.attrgetter("__name__"), operator.itemgetter(0)),
    )
    actual = sorted(
        ((k, set(v)) for k, v in toolz.groupby(type, walked_nodes).items()),
        key=toolz.compose(operator.attrgetter("__name__"), operator.itemgetter(0)),
    )
    assert actual == expected


def test_find_all_sources():
    (created_sources, _, expr) = make_expr()
    found_sources = find_all_sources(expr)
    actual = set(con._profile for con in created_sources)
    expected = set(con._profile for con in found_sources)
    assert actual == expected



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from pytest import param

import xorq as xo
import xorq.vendor.ibis.expr.types as ir
from xorq.tests.util import assert_frame_equal, assert_series_equal


@pytest.fixture(scope="module")
def flatten_data():
    return {
        "empty": {"data": [[], [], []], "type": "array<!array<!int64>>"},
        "happy": {
            "data": [[["abc"]], [["bcd"]], [["def"]]],
            "type": "array<!array<!string>>",
        },
        "nulls_only": {"data": [None, None, None], "type": "array<array<string>>"},
        "mixed_nulls": {"data": [[[]], None, [[None]]], "type": "array<array<string>>"},
    }


def test_array_column(alltypes, df):
    expr = xo.array([alltypes["double_col"], alltypes["double_col"]])
    assert isinstance(expr, ir.ArrayColumn)

    result = expr.execute()
    expected = df.apply(
        lambda row: [row["double_col"], row["double_col"]],
        axis=1,
    )
    assert_series_equal(result, expected, check_names=False)


def test_array_scalar(con):
    expr = xo.array([1.0, 2.0, 3.0])
    assert isinstance(expr, ir.ArrayScalar)

    result = con.execute(expr.name("tmp"))
    expected = np.array([1.0, 2.0, 3.0])

    assert np.array_equal(result, expected)


def test_array_repeat(con):
    expr = xo.array([1.0, 2.0]) * 2

    result = con.execute(expr.name("tmp"))
    expected = np.array([1.0, 2.0, 1.0, 2.0])

    assert np.array_equal(result, expected)


def test_array_concat(con):
    left = xo.literal([1, 2, 3])
    right = xo.literal([2, 1])
    expr = left + right
    result = con.execute(expr.name("tmp"))
    expected = np.array([1, 2, 3, 2, 1])
    assert np.array_equal(result, expected)


def test_array_concat_variadic(con):
    left = xo.literal([1, 2, 3])
    right = xo.literal([2, 1])
    expr = left.concat(right, right, right)
    result = con.execute(expr.name("tmp"))
    expected = np.array([1, 2, 3, 2, 1, 2, 1, 2, 1])
    assert np.array_equal(result, expected)


def test_array_radd_concat(con):
    left = [1]
    right = xo.literal([2])
    expr = left + right
    result = con.execute(expr.name("tmp"))
    expected = np.array([1, 2])

    assert np.array_equal(result, expected)


def test_array_length(con):
    expr = xo.literal([1, 2, 3]).length()
    assert con.execute(expr.name("tmp")) == 3


def test_list_literal(con):
    arr = [1, 2, 3]
    expr = xo.literal(arr)
    result = con.execute(expr.name("tmp"))

    assert np.array_equal(result, arr)


def test_np_array_literal(con):
    arr = np.array([1, 2, 3])
    expr = xo.literal(arr)
    result = con.execute(expr.name("tmp"))

    assert np.array_equal(result, arr)


def test_array_contains(con, array_types):
    t = array_types
    expr = t.x.contains(1)
    result = con.execute(expr)
    expected = t.x.execute().map(lambda lst: 1 in lst)
    assert_series_equal(result, expected, check_names=False)


@pytest.mark.skip(reason="failing in datafusion 34+ version")
def test_array_position(con):
    t = xo.memtable({"a": [[1], [], [42, 42], []]})
    expr = t.a.index(42)
    result = con.execute(expr)
    expected = pd.Series([-1, -1, 0, -1], dtype="object")
    assert_series_equal(result, expected, check_names=False, check_dtype=False)


def test_array_remove(con):
    t = xo.memtable({"a": [[3, 2], [], [42, 2], [2, 2], []]})
    expr = t.a.remove(2)
    result = con.execute(expr)
    expected = pd.Series([[3], [], [42], [], []], dtype="object")
    assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize(
    ("column", "expected"),
    [
        param("empty", pd.Series([[], [], []], dtype="object"), id="empty"),
        param(
            "happy", pd.Series([["abc"], ["bcd"], ["def"]], dtype="object"), id="happy"
        ),
    ],
)
def test_array_flatten(con, flatten_data, column, expected):
    data = flatten_data[column]
    t = xo.memtable({column: data["data"]}, schema={column: data["type"]})
    expr = t[column].flatten()
    result = con.execute(expr)
    assert_series_equal(
        result.reset_index(drop=True),
        expected.reset_index(drop=True),
    )


@pytest.mark.parametrize("step", [-2, -1, 1, 2])
@pytest.mark.parametrize(
    ("start", "stop"),
    [
        param(-7, -7),
        param(-7, 0),
        param(-7, 7),
        param(0, -7),
        param(0, 0),
        param(0, 7),
        param(7, -7),
        param(7, 0),
        param(7, 7),
    ],
)
def test_range_start_stop_step(con, start, stop, step):
    expr = xo.range(start, stop, step)
    result = con.execute(expr)
    assert list(result) == list(range(start, stop, step))


@pytest.mark.parametrize("n", [-2, 0, 2])
def test_range_single_argument(con, n):
    expr = xo.range(n)
    result = con.execute(expr)
    assert list(result) == list(range(n))


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        param(
            {"a": [[1, 3, 3], [], [42, 42], []]},
            [[1, 3], [], [42], []],
            id="not_null",
        ),
    ],
)
def test_array_unique(con, data, expected):
    t = xo.memtable(data)
    expr = t.a.unique()
    result = con.execute(expr)
    assert_series_equal(result, pd.Series(expected, dtype="object"))


def test_unnest_simple(array_types):
    expected = (
        array_types.execute()
        .x.explode()
        .reset_index(drop=True)
        .astype("Float64")
        .rename("tmp")
    )
    expr = array_types.x.cast("!array<float64>").unnest()
    result = expr.execute().astype("Float64").rename("tmp")

    assert_series_equal(result, expected)


def test_unnest_complex(array_types):
    df = array_types.execute()
    expr = (
        array_types.select(["grouper", "x"])
        .mutate(x=lambda t: t.x.unnest())
        .group_by("grouper")
        .aggregate(count_flat=lambda t: t.x.count())
        .order_by("grouper")
    )
    expected = (
        df[["grouper", "x"]]
        .explode("x")
        .groupby("grouper")
        .x.count()
        .rename("count_flat")
        .reset_index()
        .sort_values("grouper")
        .reset_index(drop=True)
    )
    result = expr.execute()
    assert_frame_equal(result, expected)

    # test that unnest works with to_pyarrow
    assert len(expr.to_pyarrow()) == len(result)


def test_unnest_idempotent(array_types):
    df = array_types.execute()
    expr = (
        array_types.select(
            ["scalar_column", array_types.x.cast("!array<int64>").unnest().name("x")]
        )
        .group_by("scalar_column")
        .aggregate(x=lambda t: t.x.collect())
        .order_by("scalar_column")
    )
    result = expr.execute()
    expected = (
        df[["scalar_column", "x"]]
        .assign(x=df.x.map(lambda arr: sorted(i for i in arr if not pd.isna(i))))
        .sort_values("scalar_column")
        .reset_index(drop=True)
    )
    assert_frame_equal(result, expected)


def test_unnest_no_nulls(array_types):
    df = array_types.execute()
    expr = (
        array_types.select(
            ["scalar_column", array_types.x.cast("!array<int64>").unnest().name("y")]
        )
        .filter(lambda t: t.y.notnull())
        .group_by("scalar_column")
        .aggregate(x=lambda t: t.y.collect())
        .order_by("scalar_column")
    )
    result = expr.execute()
    expected = (
        df[["scalar_column", "x"]]
        .explode("x")
        .dropna(subset=["x"])
        .groupby("scalar_column")
        .x.apply(lambda xs: [x for x in xs if x is not None])
        .reset_index()
    )
    assert_frame_equal(result, expected)


def test_unnest_default_name(array_types):
    df = array_types.execute()
    expr = (
        array_types.x.cast("!array<int64>") + xo.array([1]).cast("!array<int64>")
    ).unnest()
    assert expr.get_name().startswith("ArrayConcat(")

    result = expr.name("x").execute()
    expected = df.x.map(lambda x: x + [1]).explode("x")

    assert_series_equal(result.astype(object).fillna(pd.NA), expected.fillna(pd.NA))


@pytest.mark.parametrize(
    ("start", "stop"),
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        (-3, None),
        (-3, -1),
        (None, -3),
    ],
)
def test_array_slice(array_types, start, stop):
    expr = array_types.select(sliced=array_types.y[start:stop])
    expected = array_types.y.execute().map(lambda x: x[start:stop])
    result = expr.sliced.execute()
    assert_series_equal(result, expected)



import functools

import pandas as pd
import toolz
import xgboost as xgb
from sklearn.metrics import classification_report, confusion_matrix, roc_auc_score
from sklearn.preprocessing import OneHotEncoder

import xorq as xo
import xorq.selectors as s
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.common.utils.defer_utils import deferred_read_csv
from xorq.expr.ml import (
    deferred_fit_predict_sklearn,
    deferred_fit_transform,
    train_test_splits,
)


def fit(
    df,
    cls=functools.partial(OneHotEncoder, handle_unknown="ignore", drop="first"),
    features=slice(None),
):
    model = cls().fit(df[features])
    return model


@toolz.curry
def transform(model, df, features=slice(None)):
    names = model.get_feature_names_out()
    return pd.Series(
        (
            tuple({"key": key, "value": float(value)} for key, value in zip(names, row))
            for row in model.transform(df[features]).toarray()
        )
    )


return_type = dt.Array(dt.Struct({"key": str, "value": float}))
deferred_one_hot = deferred_fit_transform(
    fit=fit,
    transform=transform,
    return_type=return_type,
)


class XGBoostModelExplodeEncoded:
    def __init__(self, encoded_col, num_boost_round=10, params=None):
        self.encoded_col = encoded_col
        self.num_boost_round = num_boost_round
        self.params = params or {
            "max_depth": 4,
            "eta": 1,
            "objective": "binary:logistic",
            "seed": 0,
        }
        self.model = None

    def do_explode_encoded(self, X):
        X = X.drop(columns=self.encoded_col).join(
            X[self.encoded_col].apply(
                lambda lst: pd.Series({dct["key"]: dct["value"] for dct in lst})
            )
        )
        return X

    def fit(self, X, y):
        X = self.do_explode_encoded(X)
        dtrain = xgb.DMatrix(X, y)
        self.model = xgb.train(
            self.params, dtrain, num_boost_round=self.num_boost_round
        )
        return self

    def predict(self, X):
        X = self.do_explode_encoded(X)
        dmatrix = xgb.DMatrix(X)
        return self.model.predict(dmatrix)


def make_pipeline_exprs(dataset_name, target_column, predicted_col):
    ROW_NUMBER = "row_number"
    ENCODED = "encoded"

    con = xo.connect()
    train_table, test_table = (
        expr.drop(ROW_NUMBER)
        for expr in (
            deferred_read_csv(
                path=xo.options.pins.get_path(dataset_name),
                con=con,
            )
            .mutate(
                **{
                    target_column: (xo._[target_column] == "yes").cast("int"),
                    ROW_NUMBER: xo.row_number(),
                }
            )
            .pipe(
                train_test_splits,
                unique_key=ROW_NUMBER,
                test_sizes=[0.5, 0.5],
                num_buckets=2,
                random_seed=42,
            )
        )
    )

    deferred_encoder, model_udaf, deferred_encode = deferred_one_hot(
        train_table,
        features=train_table.select(s.of_type(str)).columns,
    )
    (encoded_train, encoded_test) = (
        expr.mutate(**{ENCODED: deferred_encode.on_expr})
        for expr in (train_table, test_table)
    )

    numeric_features = [
        col
        for col in encoded_train.select(s.numeric()).columns
        if col != target_column and col != target_column + "_yes"
    ]
    deferred_model, model_udaf, deferred_predict = deferred_fit_predict_sklearn(
        expr=encoded_train,
        target=target_column,
        features=numeric_features + [ENCODED],
        cls=functools.partial(XGBoostModelExplodeEncoded, encoded_col=ENCODED),
        return_type=dt.float64,
        name="xgb_prediction",
    )
    predictions = encoded_test.mutate(
        **{predicted_col: deferred_predict.on_expr(encoded_test)}
    ).drop(ENCODED)

    return {
        "encoded_train": encoded_train,
        "encoded_test": encoded_test,
        "predictions": predictions,
        "encoder": deferred_encoder,
        "model": deferred_model,
    }


(dataset_name, target_column, predicted_col) = (
    "bank-marketing",
    "deposit",
    "predicted",
)
results = make_pipeline_exprs(dataset_name, target_column, predicted_col)

predictions_df = results["predictions"].execute()
binary_predictions = (predictions_df[predicted_col] >= 0.5).astype(int)

cm = confusion_matrix(predictions_df[target_column], binary_predictions)
print("\nConfusion Matrix:")
print(f"TN: {cm[0, 0]}, FP: {cm[0, 1]}")
print(f"FN: {cm[1, 0]}, TP: {cm[1, 1]}")

auc = roc_auc_score(predictions_df[target_column], predictions_df[predicted_col])
print(f"\nAUC Score: {auc:.4f}")

print("\nClassification Report:")
print(classification_report(predictions_df[target_column], binary_predictions))



import argparse
import functools

import pandas as pd
import toolz
import xgboost as xgb
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import mean_absolute_error

import xorq as xo
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.caching import (
    ParquetStorage,
    SourceStorage,
)
from xorq.common.utils.defer_utils import deferred_read_parquet
from xorq.common.utils.import_utils import import_python
from xorq.expr.ml import (
    deferred_fit_predict,
    deferred_fit_transform_series_sklearn,
    train_test_splits,
)
from xorq.flight import (
    FlightServer,
    FlightUrl,
)
from xorq.flight.client import FlightClient


transform_port = 8765
predict_port = 8766
expected_transform_command = "execute-unbound-expr-d785a558027791af18dac689ed381d42"
expected_predict_command = "execute-unbound-expr-2f54734d557f2914929e8f0fc8784c42"


do_hackernews_fetcher_udxf = import_python(
    xo.options.pins.get_path("hackernews_lib")
).do_hackernews_fetcher_udxf
do_hackernews_sentiment_udxf = import_python(
    xo.options.pins.get_path("openai_lib")
).do_hackernews_sentiment_udxf


@toolz.curry
def fit_xgboost_model(feature_df, target_series, seed=0):
    xgb_r = xgb.XGBRegressor(
        objective="multi:softmax",
        num_class=3,
        eval_metric=mean_absolute_error,
        max_depth=6,
        # learning_rate=1,
        n_estimators=10,
        seed=seed,
    )
    X = pd.DataFrame(feature_df.squeeze().tolist())
    xgb_r.fit(X, target_series)
    return xgb_r


@toolz.curry
def predict_xgboost_model(model, df):
    return model.predict(df.squeeze().tolist())


def make_outsample_expr():
    # simulated fetch live and predict
    z = (
        xo.memtable([{"maxitem": 43346282, "n": 1000}])
        .pipe(do_hackernews_fetcher_udxf)
        .filter(xo._.text.notnull())
    )
    return z


def get_confusion_matrix(df):
    return (
        df[[target, f"{target}_predicted"]]
        .astype(int)
        .value_counts()
        .unstack(f"{target}_predicted")
    )


SENTIMENT = "sentiment"
transform_col = "title"
target = f"{SENTIMENT}_int"
do_deferred_fit_transform_tfidf = deferred_fit_transform_series_sklearn(
    col=transform_col,
    cls=TfidfVectorizer,
    return_type=dt.Array(dt.float64),
    name="transform_tfidf",
)
do_deferred_fit_predict_xgb = deferred_fit_predict(
    target=target,
    features=[f"{transform_col}_transformed"],
    fit=fit_xgboost_model,
    predict=predict_xgboost_model,
    return_type=dt.float32,
    name="predict_xgb",
)
do_train_test_split = toolz.curry(
    train_test_splits,
    unique_key="id",
    test_sizes=(0.6, 0.4),
    random_seed=42,
)


def do_fit(expr, storage=None):
    (_, _, deferred_transform) = do_deferred_fit_transform_tfidf(
        expr,
        storage=storage,
    )
    transformed = expr.mutate(
        **{f"{transform_col}_transformed": deferred_transform.on_expr}
    )
    (_, _, deferred_predict) = do_deferred_fit_predict_xgb(
        transformed,
        storage=storage,
    )
    predicted = (
        transformed
        # into_backend prevents ArrowNotImplementedError: Unsupported cast
        .into_backend(xo.connect()).mutate(
            **{f"{target}_predicted": deferred_predict.on_expr}
        )
    )
    return (predicted, deferred_transform, deferred_predict)


def do_transform_predict(expr, deferred_transform, deferred_predict):
    return (
        expr.mutate(**{f"{transform_col}_transformed": deferred_transform.on_expr})
        # into_backend prevents ArrowNotImplementedError: Unsupported cast
        .into_backend(xo.connect())
        .mutate(**{f"{target}_predicted": deferred_predict.on_expr})
    )


def do_serve(expr, deferred_transform, deferred_predict, transform_port, predict_port):
    transformed = expr.into_backend(xo.connect()).mutate(
        **{f"{transform_col}_transformed": deferred_transform.on_expr},
    )
    predicted = transformed.into_backend(xo.connect()).mutate(
        **{f"{target}_predicted": deferred_predict.on_expr},
    )
    (transform_server, transform_do_exchange) = xo.expr.relations.flight_serve(
        transformed,
        make_server=functools.partial(FlightServer, FlightUrl(port=transform_port)),
    )
    (predict_server, predict_do_exchange) = xo.expr.relations.flight_serve(
        predicted,
        make_server=functools.partial(FlightServer, FlightUrl(port=predict_port)),
    )
    return (
        transform_server,
        transform_do_exchange,
        predict_server,
        predict_do_exchange,
    )


def make_exprs():
    name = "hn-fetcher-input-large"
    con = xo.connect()
    storage = ParquetStorage(source=con)
    # pg.postgres.connect_env().create_catalog("caching")
    pg = xo.postgres.connect_env(database="caching")

    (train_expr, test_expr) = (
        deferred_read_parquet(
            con,
            xo.options.pins.get_path(name),
            name,
        )
        .pipe(do_hackernews_fetcher_udxf)
        .filter(xo._.text.notnull())
        .cache(storage=SourceStorage(pg))
        .pipe(do_hackernews_sentiment_udxf, con=con)
        .cache(storage=SourceStorage(pg))
        .cache(storage=ParquetStorage(con))
        .filter(~xo._[SENTIMENT].contains("ERROR"))
        .mutate(
            **{
                target: (
                    xo._[SENTIMENT]
                    .cases(
                        (
                            ("POSITIVE", 2),
                            ("NEUTRAL", 1),
                            ("NEGATIVE", 0),
                        )
                    )
                    .cast(int)
                ),
            }
        )
        .pipe(do_train_test_split)
    )
    (train_predicted, deferred_transform, deferred_predict) = do_fit(
        train_expr, storage
    )
    test_predicted = do_transform_predict(
        test_expr, deferred_transform, deferred_predict
    )
    return (train_predicted, test_predicted, deferred_transform, deferred_predict)


def run():
    (train_predicted, test_predicted, deferred_transform, deferred_predict) = (
        make_exprs()
    )
    print(train_predicted.execute().pipe(get_confusion_matrix))
    print(test_predicted.execute().pipe(get_confusion_matrix))
    return (train_predicted, test_predicted, deferred_transform, deferred_predict)


def serve(transform_port, predict_port):
    (_, _, deferred_transform, deferred_predict) = make_exprs()
    z = make_outsample_expr()
    (transform_server, transform_do_exchange, predict_server, predict_do_exchange) = (
        do_serve(z, deferred_transform, deferred_predict, transform_port, predict_port)
    )
    validate_commands(transform_do_exchange, predict_do_exchange)
    return (
        transform_server,
        transform_do_exchange,
        predict_server,
        predict_do_exchange,
    )


def predict(transform_port, predict_port):
    transform_client = FlightClient(port=transform_port)
    transform_do_exchange = toolz.curry(
        transform_client.do_exchange, expected_transform_command
    )
    predict_client = FlightClient(port=predict_port)
    predict_do_exchange = toolz.curry(
        predict_client.do_exchange, expected_predict_command
    )
    predicted = predict_do_exchange(
        transform_do_exchange(make_outsample_expr().to_pyarrow_batches())[1]
    )[1].read_pandas()
    print(predicted)
    return predicted


def validate_commands(transform_do_exchange, predict_do_exchange):
    (transform_command, predict_command) = (
        do_exchange.args[1]
        for do_exchange in (transform_do_exchange, predict_do_exchange)
    )
    assert transform_command == expected_transform_command
    assert predict_command == expected_predict_command


def parse_args(override=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("command", choices=("run", "serve", "predict"))
    parser.add_argument("-t", "--transform-port", default=transform_port, type=int)
    parser.add_argument("-p", "--predict-port", default=predict_port, type=int)
    args = parser.parse_args(override)
    return args


if __name__ == "__main__":
    args = parse_args()
    match args.command:
        case "run":
            (train_predicted, test_predicted, deferred_transform, deferred_predict) = (
                run()
            )
        case "serve":
            (
                transform_server,
                transform_do_exchange,
                predict_server,
                predict_do_exchange,
            ) = serve(args.transform_port, args.predict_port)
        case "predict":
            predicted = predict(args.transform_port, args.predict_port)
        case _:
            raise ValueError



from __future__ import annotations

import pickle

import numpy as np
import pandas as pd
import pytest
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LinearRegression

import xorq as xo
import xorq.expr.datatypes as dt
from xorq import memtable
from xorq.caching import ParquetStorage
from xorq.common.utils.defer_utils import (
    deferred_read_parquet,
)
from xorq.expr.ml import _calculate_bounds
from xorq.ml import (
    deferred_fit_predict_sklearn,
    deferred_fit_transform_series_sklearn,
    make_quickgrove_udf,
)
from xorq.tests.util import assert_frame_equal


def test_train_test_splits_intersections():
    # This is testing the base case where a single float becomes ( 1-test_size , test_size ) proportion
    # Check counts and overlaps in train and test dataset
    N = 10000
    test_size = [0.1, 0.2, 0.7]

    # init table
    table = memtable([(i, "val") for i in range(N)], columns=["key1", "val"])
    results = [
        r
        for r in xo.train_test_splits(
            table,
            unique_key="key1",
            test_sizes=test_size,
            num_buckets=N,
            random_seed=42,
        )
    ]

    # make sure all splits mutually exclusive
    # These are all  a \ b  U  a intersect b  where b are the other splits
    element1 = results[0]
    complement1 = results[1].union(results[2])

    element2 = results[1]
    complement2 = results[0].union(results[2])

    element3 = results[2]
    complement3 = results[0].union(results[1])

    assert element1.union(complement1).join(table, how="anti").count().execute() == 0
    assert (
        element1.join(complement1, element1.key1 == complement1.key1).count().execute()
        == 0
    )

    assert element2.union(complement2).join(table, how="anti").count().execute() == 0
    assert (
        element2.join(complement2, element2.key1 == complement2.key1).count().execute()
        == 0
    )

    assert element3.union(complement3).join(table, how="anti").count().execute() == 0
    assert (
        element3.join(complement3, element3.key1 == complement3.key1).count().execute()
        == 0
    )


@pytest.mark.xfail
def test_train_test_split():
    # This is testing the base case where a single float becomes ( 1-test_size , test_size ) proportion
    # Check counts and overlaps in train and test dataset
    N = 100
    test_size = 0.25

    # init table
    table = memtable([(i, "val") for i in range(N)], columns=["key1", "val"])
    train_table, test_table = xo.train_test_splits(
        table, unique_key="key1", test_sizes=test_size, num_buckets=N, random_seed=42
    )

    # These values are for seed 42
    assert train_table.count().execute() == 75
    assert test_table.count().execute() == 25
    assert set(train_table.columns) == set(table.columns)
    assert set(test_table.columns) == set(table.columns)
    # make sure data unioned together is itself
    assert train_table.union(test_table).join(table, how="semi").count().execute() == N

    # Check reproducibility
    reproduced_train_table, reproduced_test_table = xo.train_test_splits(
        table, unique_key="key1", test_sizes=test_size, num_buckets=N, random_seed=42
    )
    assert_frame_equal(train_table.execute(), reproduced_train_table.execute())
    assert_frame_equal(test_table.execute(), reproduced_test_table.execute())

    # make sure it could generate different data with different random_seed
    different_train_table, different_test_table = xo.train_test_splits(
        table, unique_key="key1", test_sizes=test_size, num_buckets=N, random_seed=0
    )
    assert not train_table.execute().equals(different_train_table.execute())
    assert not test_table.execute().equals(different_test_table.execute())


def test_train_test_split_invalid_test_size():
    table = memtable({"key": [1, 2, 3]})
    with pytest.raises(ValueError, match="test size should be a float between 0 and 1"):
        xo.train_test_splits(table, unique_key="key", test_sizes=1.5)
    with pytest.raises(ValueError, match="test size should be a float between 0 and 1"):
        xo.train_test_splits(table, unique_key="key", test_sizes=-0.5)


def test_train_test_split_invalid_num_buckets_type():
    table = memtable({"key": [1, 2, 3]})
    with pytest.raises(ValueError, match="num_buckets must be an integer"):
        xo.train_test_splits(table, unique_key="key", test_sizes=0.5, num_buckets=10.5)


def test_train_test_split_invalid_num_buckets_value():
    table = memtable({"key": [1, 2, 3]})
    with pytest.raises(
        ValueError, match="num_buckets = 1 places all data into training set"
    ):
        xo.train_test_splits(table, unique_key="key", test_sizes=0.5, num_buckets=1)


def test_train_test_split_multiple_keys():
    data = {
        "key1": range(100),
        "key2": [chr(i % 26 + 65) for i in range(100)],  # A, B, C, ...
        "value": [i % 3 for i in range(100)],
    }
    table = memtable(data)
    train_table, test_table = xo.train_test_splits(
        table,
        unique_key=["key1", "key2"],
        test_sizes=0.25,
        num_buckets=10,
        random_seed=99,
    )
    assert train_table.union(test_table).join(table, how="anti").count().execute() == 0


@pytest.mark.xfail
def test_train_test_splits_deterministic_with_seed():
    table = memtable({"key": range(100), "value": range(100)})
    test_sizes = [0.4, 0.6]

    splits1 = list(
        xo.train_test_splits(table, "key", test_sizes, random_seed=123, num_buckets=10)
    )
    splits2 = list(
        xo.train_test_splits(table, "key", test_sizes, random_seed=123, num_buckets=10)
    )

    for s1, s2 in zip(splits1, splits2):
        assert_frame_equal(s1.execute(), s2.execute())


def test_train_test_splits_invalid_test_sizes():
    table = memtable({"key": [1, 2, 3], "value": [4, 5, 6]})
    with pytest.raises(ValueError, match="Test size must be float."):
        next(xo.train_test_splits(table, "key", ["a", "b"]))
    with pytest.raises(
        ValueError, match="test size should be a float between 0 and 1."
    ):
        next(xo.train_test_splits(table, "key", [-0.1, 0.5]))


def test_train_test_splits_must_sum_one():
    table = memtable({"key": [1, 2, 3], "value": [4, 5, 6]})
    with pytest.raises(ValueError, match="Test sizes must sum to 1"):
        next(xo.train_test_splits(table, "key", [0.1, 0.5]))


@pytest.mark.parametrize(
    "test_sizes",
    ((1 / n,) * n for n in range(2, 100, 5)),
)
def test_approx_sum(test_sizes):
    _calculate_bounds(test_sizes)


def test_calculate_bounds():
    test_sizes = [0.2, 0.3, 0.5]
    expected_bounds = ((0.0, 0.2), (0.2, 0.5), (0.5, 1.0))
    assert _calculate_bounds(test_sizes) == expected_bounds


def test_train_test_splits_num_buckets_gt_one():
    table = memtable({"key": range(100), "value": range(100)})
    test_sizes = [0.4, 0.6]
    with pytest.raises(
        ValueError,
        match="num_buckets = 1 places all data into training set. For any integer x  >=0 , x mod 1 = 0 . ",
    ):
        next(
            xo.train_test_splits(
                table, "key", test_sizes, random_seed=123, num_buckets=1
            )
        )


@pytest.mark.parametrize(
    "connect_method",
    (
        xo.connect,
        xo.duckdb.connect,
        xo.postgres.connect_env,
        pytest.param(
            xo.datafusion.connect,
            marks=pytest.mark.xfail(
                reason="Compilation rule for 'Hash' operation is not define"
            ),
        ),
    ),
)
def test_train_test_splits_intersections_parameterized_pass(connect_method):
    # This is testing the base case where a single float becomes ( 1-test_size , test_size ) proportion
    # Check counts and overlaps in train and test dataset
    N = 10000
    test_size = [0.1, 0.2, 0.7]

    # create test table for backend
    test_df = pd.DataFrame([(i, "val") for i in range(N)], columns=["key1", "val"])
    con = connect_method()
    test_table_name = f"{con.name}_test_df"
    con.create_table(test_table_name, test_df, temp=con.name == "postgres")

    table = con.table(test_table_name)

    results = [
        r
        for r in xo.train_test_splits(
            table,
            unique_key="key1",
            test_sizes=test_size,
            num_buckets=N,
            random_seed=42,
        )
    ]

    # make sure all splits mutually exclusive
    # These are all  a \ b  U  a intersect b  where b are the other splits
    element1 = results[0]
    complement1 = results[1].union(results[2])

    element2 = results[1]
    complement2 = results[0].union(results[2])

    element3 = results[2]
    complement3 = results[0].union(results[1])

    assert element1.union(complement1).join(table, how="anti").count().execute() == 0
    assert (
        element1.join(complement1, element1.key1 == complement1.key1).count().execute()
        == 0
    )

    assert element2.union(complement2).join(table, how="anti").count().execute() == 0
    assert (
        element2.join(complement2, element2.key1 == complement2.key1).count().execute()
        == 0
    )

    assert element3.union(complement3).join(table, how="anti").count().execute() == 0
    assert (
        element3.join(complement3, element3.key1 == complement3.key1).count().execute()
        == 0
    )
    con.drop_table(test_table_name)


@pytest.mark.parametrize(
    "connect_method",
    (
        xo.connect,
        xo.duckdb.connect,
        xo.postgres.connect_env,
        pytest.param(
            xo.datafusion.connect,
            marks=pytest.mark.xfail(
                reason="Compilation rule for 'Hash' operation is not define"
            ),
        ),
    ),
)
@pytest.mark.parametrize("n", (2, 8, 32))
@pytest.mark.parametrize("name", ("split", "other"))
def test_calc_split_column(connect_method, n, name):
    N = 10000
    test_sizes = (1 / n,) * n
    unique_key = "key1"

    # create test table for backend
    test_df = pd.DataFrame([(i, "val") for i in range(N)], columns=[unique_key, "val"])
    con = connect_method()
    test_table_name = f"{con.name}_test_df"
    con.create_table(test_table_name, test_df, temp=con.name == "postgres")

    table = con.table(test_table_name)
    expr = (
        table.mutate(
            xo.calc_split_column(
                table,
                unique_key=unique_key,
                test_sizes=test_sizes,
                random_seed=42,
                name=name,
            )
        )[name]
        .value_counts()
        .order_by(xo.asc(name))
    )
    df = xo.execute(expr)
    assert tuple(df[name].values) == tuple(range(n))
    assert df[f"{name}_count"].sum() == N


def test_make_quickgrove_udf_predictions(feature_table, float_model_path):
    """quickgrove UDF predictions should match expected values"""
    predict_udf = make_quickgrove_udf(float_model_path)
    result = feature_table.mutate(pred=predict_udf.on_expr).execute()

    np.testing.assert_almost_equal(
        result["pred"].values, result["expected_pred"].values, decimal=3
    )


def test_make_quickgrove_udf_signature(float_model_path):
    """quickgrove UDF should have correct signature with float64 inputs, float32 output"""
    predict_fn = make_quickgrove_udf(float_model_path)

    assert predict_fn.__signature__.return_annotation == dt.float32
    assert all(
        p.annotation == dt.float64 for p in predict_fn.__signature__.parameters.values()
    )
    assert predict_fn.__name__ == "pretrained_model"


def test_make_quickgrove_udf_mixed_features(mixed_model_path):
    """quickgrove UDF should support int64 and boolean feature types"""
    predict_fn = make_quickgrove_udf(mixed_model_path)
    assert "i" in predict_fn.model.feature_types


def test_make_quickgrove_udf__repr(mixed_model_path):
    """quickgrove UDF repr should include model metadata"""
    predict_fn = make_quickgrove_udf(mixed_model_path)
    repr_str = repr(predict_fn)

    expected_info = [
        "Total number of trees:",
        "Average tree depth:",
        "Max tree depth:",
        "Total number of nodes:",
        "Model path:",
        "Signature:",
    ]

    for info in expected_info:
        assert info in repr_str


def test_quickgrove_hyphen_name(feature_table, hyphen_model_path):
    assert "-" in hyphen_model_path.name
    with pytest.raises(
        ValueError,
        match="The argument model_name was None and the name extracted from the path is not a valid Python identifier",
    ):
        make_quickgrove_udf(hyphen_model_path)

    with pytest.raises(
        ValueError, match="The argument model_name is not a valid Python identifier"
    ):
        make_quickgrove_udf(hyphen_model_path, "diamonds-model")

    predict_udf = make_quickgrove_udf(hyphen_model_path, model_name="diamonds_model")
    result = feature_table.mutate(pred=predict_udf.on_expr).execute()

    np.testing.assert_almost_equal(
        result["pred"].values, result["expected_pred"].values, decimal=3
    )


def make_data():
    X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
    # y = 1 * x_0 + 2 * x_1 + 3
    y = np.dot(X, np.array([1, 2])) + 3
    df = pd.DataFrame(np.hstack((X, y[:, np.newaxis]))).rename(
        columns=lambda x: chr(x + ord("a"))
    )
    (*features, target) = df.columns
    return (df, features, target)


deferred_linear_regression = deferred_fit_predict_sklearn(
    cls=LinearRegression, return_type=dt.float64
)


def test_deferred_fit_predict_linear_regression(tmp_path):
    con = xo.connect()
    (df, features, target) = make_data()
    t = con.register(df, "t")

    # uncached run
    (computed_kwargs_expr, _, predict_expr_udf) = deferred_linear_regression(
        t, target, features
    )
    model = computed_kwargs_expr.execute()
    predicted = t.mutate(predict_expr_udf.on_expr(t)).execute()

    # cached run
    storage = ParquetStorage(path=tmp_path, source=con)
    (computed_kwargs_expr, _, predict_expr_udf) = deferred_linear_regression(
        t, target, features, storage=storage
    )
    ((cached_model,),) = computed_kwargs_expr.execute().values
    cached_predicted = t.mutate(predict_expr_udf.on_expr(t)).execute()

    assert predicted.equals(cached_predicted)
    np.testing.assert_almost_equal(
        pickle.loads(model).coef_, pickle.loads(cached_model).coef_
    )


def test_deferred_fit_predict_linear_regression_multi_into_backend():
    con = xo.connect()
    (df, features, target) = make_data()
    t = (
        con.register(df, "t")
        .into_backend(xo.connect())[lambda t: t[features[0]] > 0]
        .into_backend(xo.connect())
    )

    (_, _, predict_expr_udf) = deferred_linear_regression(t, target, features)
    predicted = t.mutate(predict_expr_udf.on_expr(t)).execute()
    assert not predicted.empty


def test_deferred_fit_transform_series_sklearn():
    cls = TfidfVectorizer
    transform_key = "transformed"
    col = "title"
    deferred_fit_transform_tfidf = deferred_fit_transform_series_sklearn(
        col=col, cls=cls, return_type=dt.Array(dt.float64)
    )
    con = xo.connect()
    train, test = deferred_read_parquet(
        con,
        xo.options.pins.get_path("hn-data-small.parquet"),
        "fetcher-input",
    ).pipe(
        xo.train_test_splits,
        unique_key="id",
        test_sizes=(0.9, 0.1),
    )
    (_, _, deferred_transform) = deferred_fit_transform_tfidf(
        train,
    )

    from_sklearn = test.execute().assign(
        **{
            transform_key: lambda t: (
                cls().fit(train.execute()[col]).transform(t[col]).toarray().tolist()
            )
        }
    )
    from_xo = test.mutate(**{transform_key: deferred_transform.on_expr}).execute()
    actual = from_xo[transform_key].apply(pd.Series)
    expected = from_sklearn[transform_key].apply(pd.Series)
    assert actual.equals(expected)


@pytest.mark.parametrize(
    "method",
    [
        "train_test_splits",
        "deferred_fit_predict",
        "deferred_fit_predict_sklearn",
        "deferred_fit_transform",
        "deferred_fit_transform_series_sklearn",
        "make_quickgrove_udf",
        "rewrite_quickgrove_expr",
    ],
)
def test_top_level_ml(method):
    assert hasattr(xo, method)
    assert hasattr(xo.ml, method)



from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def root_dir():
    return Path(__file__).absolute().parents[2]


@pytest.fixture(scope="session")
def parquet_dir(root_dir):
    data_dir = root_dir / "ci" / "ibis-testing-data" / "parquet"
    return data_dir



import functools
import pickle
from typing import Any

import pyarrow as pa
import toolz

import xorq.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor.ibis.common.annotations import Argument
from xorq.vendor.ibis.common.collections import FrozenDict
from xorq.vendor.ibis.expr.operations import Namespace
from xorq.vendor.ibis.expr.operations.udf import (
    AggUDF,
    InputType,
    ScalarUDF,
    _make_udf_name,
    _wrap,
    scalar,
)
from xorq.vendor.ibis.expr.operations.udf import (
    agg as _agg,
)


@toolz.curry
def wrap_model(value, model_key="model"):
    return pickle.dumps({model_key: value})


unwrap_model = pickle.loads


class ExprScalarUDF(ScalarUDF):
    @property
    def computed_kwargs_expr(self):
        # must push the expr into __config__ so that it doesn't get turned into a window function
        return self.__config__["computed_kwargs_expr"]

    @property
    def post_process_fn(self):
        return self.__config__["post_process_fn"]

    @property
    def schema(self):
        return self.__config__["schema"]

    def on_expr(self, e, **kwargs):
        # rebind deferred_model (computed_kwargs_expr) to a new expr
        return type(self)(*(e[c] for c in self.schema), **kwargs)


@toolz.curry
def make_pandas_expr_udf(
    computed_kwargs_expr,
    fn,
    schema,
    return_type=dt.binary,
    database=None,
    catalog=None,
    name=None,
    *,
    post_process_fn=unwrap_model,
    **kwargs,
):
    import pandas as pd

    def fn_from_arrays(*arrays, computed_arg=None, **kwargs):
        if computed_arg is None:
            raise ValueError(
                "Caller must bind computed_arg to the output of computed_kwargs_expr"
            )
        df = pd.DataFrame(
            {name: array.to_pandas() for (name, array) in zip(schema, arrays)}
        )
        return pa.Array.from_pandas(
            fn(computed_arg, df, **kwargs), type=return_type.to_pyarrow()
        )

    name = name if name is not None else _make_udf_name(fn)
    bases = (ExprScalarUDF,)
    fields = {
        arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (arg_name, typ) in schema.items()
    }
    meta = {
        "dtype": return_type,
        "__input_type__": InputType.PYARROW,
        "__func__": property(
            fget=lambda _, fn_from_arrays=fn_from_arrays: fn_from_arrays
        ),
        # valid config keys: computed_kwargs_expr, post_process_fn, volatility
        "__config__": FrozenDict(
            computed_kwargs_expr=computed_kwargs_expr,
            post_process_fn=post_process_fn,
            schema=schema,
            **kwargs,
        ),
        "__udf_namespace__": Namespace(database=database, catalog=catalog),
        "__module__": fn.__module__,
        # FIXME: determine why this fails with case mismatch by default
        "__func_name__": name,
    }
    kwds = {
        **fields,
        **meta,
    }

    node = type(
        name,
        bases,
        kwds,
    )

    # FIXME: enable working with deferred like _wrap enables
    @functools.wraps(fn)
    def construct(*args: Any, **kwargs: Any) -> ir.Value:
        return node(*args, **kwargs).to_expr()

    def on_expr(e, **kwargs):
        return construct(*(e[c] for c in schema), **kwargs)

    construct.on_expr = on_expr
    construct.fn = fn
    return construct


@toolz.curry
def make_pandas_udf(
    fn, schema, return_type, database=None, catalog=None, name=None, **kwargs
):
    import pandas as pd

    from xorq.vendor.ibis.expr.operations.udf import ScalarUDF

    def fn_from_arrays(*arrays):
        df = pd.DataFrame(
            {name: array.to_pandas() for (name, array) in zip(schema, arrays)}
        )
        return pa.Array.from_pandas(fn(df), type=return_type.to_pyarrow())

    name = name if name is not None else _make_udf_name(fn)
    bases = (ScalarUDF,)
    fields = {
        arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (arg_name, typ) in schema.items()
    }
    meta = {
        "dtype": return_type,
        "__input_type__": InputType.PYARROW,
        "__func__": property(
            fget=lambda _, fn_from_arrays=fn_from_arrays: fn_from_arrays
        ),
        # valid config keys: volatility
        "__config__": FrozenDict(**kwargs),
        "__udf_namespace__": Namespace(database=database, catalog=catalog),
        "__module__": fn.__module__,
        # FIXME: determine why this fails with case mismatch by default
        "__func_name__": name,
    }
    kwds = {
        **fields,
        **meta,
    }

    node = type(
        name,
        bases,
        kwds,
    )

    # FIXME: enable working with deferred like _wrap enables
    @functools.wraps(fn)
    def construct(*args: Any, **kwargs: Any) -> ir.Value:
        return node(*args, **kwargs).to_expr()

    def on_expr(e, **kwargs):
        return construct(*(e[c] for c in schema), **kwargs)

    construct.on_expr = on_expr
    construct.fn = fn
    return construct


class agg(_agg):
    __slots__ = ()

    _base = AggUDF

    @classmethod
    def pyarrow(
        cls,
        fn=None,
        name=None,
        signature=None,
        **kwargs,
    ):
        result = _wrap(
            cls._make_wrapper,
            InputType.PYARROW,
            fn,
            name=name,
            signature=signature,
            **kwargs,
        )
        return result

    @classmethod
    @toolz.curry
    def pandas_df(
        cls,
        fn,
        schema,
        return_type,
        database=None,
        catalog=None,
        name=None,
        **kwargs,
    ):
        import pandas as pd

        def fn_from_arrays(*arrays):
            df = pd.DataFrame(
                {name: array.to_pandas() for (name, array) in zip(schema, arrays)}
            )
            return fn(df)

        name = name if name is not None else _make_udf_name(fn)
        bases = (cls._base,)
        fields = {
            arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
            for (arg_name, typ) in schema.items()
        }
        meta = {
            "dtype": return_type,
            "__input_type__": InputType.PYARROW,
            "__func__": property(
                fget=lambda _, fn_from_arrays=fn_from_arrays: fn_from_arrays
            ),
            # valid config keys: volatility
            "__config__": FrozenDict(**kwargs),
            "__udf_namespace__": Namespace(database=database, catalog=catalog),
            "__module__": fn.__module__,
            # FIXME: determine why this fails with case mismatch by default
            "__func_name__": name,
        }
        kwds = {
            **fields,
            **meta,
        }

        node = type(
            name,
            bases,
            kwds,
        )

        # FIXME: enable working with deferred like _wrap enables
        @functools.wraps(fn)
        def construct(*args: Any, **kwargs: Any) -> ir.Value:
            return node(*args, **kwargs).to_expr()

        def on_expr(e, **kwargs):
            return construct(*(e[c] for c in schema), **kwargs)

        construct.on_expr = on_expr
        construct.fn = fn
        return construct


def arbitrate_evaluate(
    uses_window_frame=False,
    supports_bounded_execution=False,
    include_rank=False,
    **config_kwargs,
):
    match (uses_window_frame, supports_bounded_execution, include_rank):
        case (False, False, False):
            return "evaluate_all"
        case (False, True, False):
            return "evaluate"
        case (False, _, True):
            return "evaluate_all_with_rank"
        case (True, _, _):
            return "evaluate"
        case _:
            raise RuntimeError


@toolz.curry
def pyarrow_udwf(
    fn,
    schema,
    return_type,
    name=None,
    namespace=Namespace(database=None, catalog=None),
    base=AggUDF,
    **config_kwargs,
):
    fields = {
        arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (arg_name, typ) in schema.items()
    }
    # which_evaluate = arbitrate_evaluate(**config_kwargs)
    name = name or fn.__name__
    meta = {
        "dtype": return_type,
        "__input_type__": InputType.PYARROW,
        "__func__": property(fget=toolz.functoolz.return_none),
        "__config__": FrozenDict(
            input_types=tuple(el.type for el in schema.to_pyarrow()),
            return_type=return_type.to_pyarrow(),
            name=name,
            **config_kwargs,
            # assert which_evaluate in ("evaluate", "evaluate_all", "evaluate_all_with_rank")
            # **{which_evaluate: fn},
            **{
                which_evaluate: fn
                for which_evaluate in (
                    "evaluate",
                    "evaluate_all",
                    "evaluate_all_with_rank",
                )
            },
        ),
        "__udf_namespace__": namespace,
        "__module__": __name__,
        "__func_name__": name,
    }
    node = type(
        name,
        (base,),
        {
            **fields,
            **meta,
        },
    )

    def construct(*args: Any, **kwargs: Any) -> ir.Value:
        return node(*args, **kwargs).to_expr()

    def on_expr(e, **kwargs):
        return construct(*(e[c] for c in schema), **kwargs)

    construct.on_expr = on_expr
    return construct


__all__ = [
    "pyarrow_udwf",
    "make_pandas_expr_udf",
    "make_pandas_udf",
    "scalar",
    "agg",
]



import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import xorq as xo
from xorq.common.utils.aws_utils import (
    connection_is_set,
    make_s3_credentials_defaults,
)
from xorq.vendor import ibis


FIXTURES_DIR = Path(__file__).parent / "fixtures"

TEST_TABLES = {
    "functional_alltypes": ibis.schema(
        {
            "id": "int32",
            "bool_col": "boolean",
            "tinyint_col": "int8",
            "smallint_col": "int16",
            "int_col": "int32",
            "bigint_col": "int64",
            "float_col": "float32",
            "double_col": "float64",
            "date_string_col": "string",
            "string_col": "string",
            "timestamp_col": "timestamp",
            "year": "int32",
            "month": "int32",
        }
    ),
    "diamonds": ibis.schema(
        {
            "carat": "float64",
            "cut": "string",
            "color": "string",
            "clarity": "string",
            "depth": "float64",
            "table": "float64",
            "price": "int64",
            "x": "float64",
            "y": "float64",
            "z": "float64",
        }
    ),
    "batting": ibis.schema(
        {
            "playerID": "string",
            "yearID": "int64",
            "stint": "int64",
            "teamID": "string",
            "lgID": "string",
            "G": "int64",
            "AB": "int64",
            "R": "int64",
            "H": "int64",
            "X2B": "int64",
            "X3B": "int64",
            "HR": "int64",
            "RBI": "int64",
            "SB": "int64",
            "CS": "int64",
            "BB": "int64",
            "SO": "int64",
            "IBB": "int64",
            "HBP": "int64",
            "SH": "int64",
            "SF": "int64",
            "GIDP": "int64",
        }
    ),
    "awards_players": ibis.schema(
        {
            "playerID": "string",
            "awardID": "string",
            "yearID": "int64",
            "lgID": "string",
            "tie": "string",
            "notes": "string",
        }
    ),
    "astronauts": ibis.schema(
        {
            "id": "int64",
            "number": "int64",
            "nationwide_number": "int64",
            "name": "string",
            "original_name": "string",
            "sex": "string",
            "year_of_birth": "int64",
            "nationality": "string",
            "military_civilian": "string",
            "selection": "string",
            "year_of_selection": "int64",
            "mission_number": "int64",
            "total_number_of_missions": "int64",
            "occupation": "string",
            "year_of_mission": "int64",
            "mission_title": "string",
            "ascend_shuttle": "string",
            "in_orbit": "string",
            "descend_shuttle": "string",
            "hours_mission": "float64",
            "total_hrs_sum": "float64",
            "field21": "int64",
            "eva_hrs_mission": "float64",
            "total_eva_hrs": "float64",
        }
    ),
}

array_types_df = pd.DataFrame(
    [
        (
            [np.int64(1), 2, 3],
            ["a", "b", "c"],
            [1.0, 2.0, 3.0],
            "a",
            1.0,
            [[], [np.int64(1), 2, 3], None],
        ),
        (
            [4, 5],
            ["d", "e"],
            [4.0, 5.0],
            "a",
            2.0,
            [],
        ),
        (
            [6, None],
            ["f", None],
            [6.0, np.nan],
            "a",
            3.0,
            [None, [], None],
        ),
        (
            [None, 1, None],
            [None, "a", None],
            [],
            "b",
            4.0,
            [[1], [2], [], [3, 4, 5]],
        ),
        (
            [2, None, 3],
            ["b", None, "c"],
            np.nan,
            "b",
            5.0,
            None,
        ),
        (
            [4, None, None, 5],
            ["d", None, None, "e"],
            [4.0, np.nan, np.nan, 5.0],
            "c",
            6.0,
            [[1, 2, 3]],
        ),
    ],
    columns=[
        "x",
        "y",
        "z",
        "grouper",
        "scalar_column",
        "multi_dim",
    ],
)

have_s3_credentials = connection_is_set(make_s3_credentials_defaults())


def pytest_runtest_setup(item):
    if any(mark.name == "s3" for mark in item.iter_markers()):
        if not have_s3_credentials:
            pytest.skip("cannot run s3 tests without s3 credentials")


@pytest.fixture(scope="session")
def data_dir():
    root = Path(__file__).absolute().parents[3]
    data_dir = root / "ci" / "ibis-testing-data"
    return data_dir


@pytest.fixture(scope="session")
def ddl_file():
    root = Path(__file__).absolute().parents[3]
    ddl_dir = root / "db" / "datafusion.sql"
    return ddl_dir


def statements(ddl_file: Path):
    return (
        statement
        for statement in map(str.strip, ddl_file.read_text().split(";"))
        if statement
    )


@pytest.fixture(scope="session")
def con(data_dir, ddl_file):
    conn = xo.connect()
    parquet_dir = data_dir / "parquet"
    conn.read_parquet(
        parquet_dir / "functional_alltypes.parquet", "functional_alltypes"
    )
    conn.read_parquet(parquet_dir / "batting.parquet", "batting")
    conn.read_parquet(parquet_dir / "diamonds.parquet", "diamonds")
    conn.read_parquet(parquet_dir / "astronauts.parquet", "astronauts")
    conn.read_parquet(parquet_dir / "awards_players.parquet", "awards_players")

    conn.create_table("array_types", array_types_df)

    if ddl_file.is_file() and ddl_file.name.endswith(".sql"):
        for statement in statements(ddl_file):
            with conn._safe_raw_sql(statement):  # noqa
                pass

    return conn


@pytest.fixture(scope="session")
def functional_alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def batting(con):
    return con.table("batting")


@pytest.fixture(scope="session")
def batting_df(batting):
    return batting.execute()


@pytest.fixture(scope="session")
def awards_players(con):
    return con.table("awards_players")


@pytest.fixture(scope="session")
def awards_players_df(awards_players):
    return awards_players.execute(limit=None)


@pytest.fixture(scope="session")
def sorted_df(df):
    return df.sort_values("id").reset_index(drop=True)


@pytest.fixture(scope="session")
def diamonds(con):
    return con.table("diamonds")


@pytest.fixture(scope="session")
def array_types(con):
    return con.table("array_types")


@pytest.fixture(scope="session")
def struct(con):
    return con.table("structs")


@pytest.fixture(scope="session")
def struct_df(struct):
    return struct.execute()


@pytest.fixture(scope="session")
def float_model_path():
    """Diamonds model with all float features"""
    return FIXTURES_DIR / "pretrained_model.json"


@pytest.fixture(scope="session")
def mixed_model_path():
    return FIXTURES_DIR / "pretrained_model_mixed.json"


@pytest.fixture(scope="session")
def hyphen_model_path(tmp_path_factory):
    model_path = FIXTURES_DIR / "pretrained_model.json"
    fn = tmp_path_factory.mktemp("data") / "diamonds-model.json"
    shutil.copy(model_path, fn)
    return fn


@pytest.fixture(scope="session")
def mixed_feature_table():
    con = xo.connect()
    df = pd.DataFrame(
        {
            "carat": [0.23, 0.21, 0.23],
            "depth": [61.5, None, 56.9],
            "table": [55.0, 61.0, 65.0],
            "x": [3.95, 3.89, 4.05],
            "y": [3.98, 3.84, 4.07],
            "z": [2.43, 2.31, 2.31],
            "cut_good": [False, False, True],
            "cut_ideal": [True, False, False],
            "cut_premium": [False, True, False],
            "cut_very_good": [False, False, False],
            "color_e": [True, True, True],
            "color_f": [False, False, False],
            "color_g": [False, False, False],
            "color_h": [False, False, False],
            "color_i": [False, False, False],
            "color_j": [False, False, False],
            "clarity_if": [False, False, False],
            "clarity_si1": [False, True, False],
            "clarity_si2": [True, False, False],
            "clarity_vs1": [False, False, True],
            "clarity_vs2": [False, False, False],
            "clarity_vvs1": [False, False, False],
            "clarity_vvs2": [False, False, False],
            "target": [326, 326, 327],
            "expected_pred": [472.00235, 580.98920, 480.31976],
        }
    )
    return con.create_table("mixed_table", df)


@pytest.fixture(scope="session")
def feature_table():
    con = xo.connect()
    df = pd.DataFrame(
        {
            "carat": [0.23, 0.21, 0.23],
            "depth": [61.5, None, 56.9],
            "table": [55.0, 61.0, 65.0],
            "x": [3.95, 3.89, 4.05],
            "y": [3.98, 3.84, 4.07],
            "z": [2.43, 2.31, 2.31],
            "cut_good": [0.0, 0.0, 1.0],
            "cut_ideal": [1.0, 0.0, 0.0],
            "cut_premium": [0.0, 1.0, 0.0],
            "cut_very_good": [0.0, 0.0, 0.0],
            "color_e": [1.0, 1.0, 1.0],
            "color_f": [0.0, 0.0, 0.0],
            "color_g": [0.0, 0.0, 0.0],
            "color_h": [0.0, 0.0, 0.0],
            "color_i": [0.0, 0.0, 0.0],
            "color_j": [0.0, 0.0, 0.0],
            "clarity_if": [0.0, 0.0, 0.0],
            "clarity_si1": [0.0, 1.0, 0.0],
            "clarity_si2": [1.0, 0.0, 0.0],
            "clarity_vs1": [0.0, 0.0, 1.0],
            "clarity_vs2": [0.0, 0.0, 0.0],
            "clarity_vvs1": [0.0, 0.0, 0.0],
            "clarity_vvs2": [0.0, 0.0, 0.0],
            "target": [326, 326, 327],
            "expected_pred": [472.00235, 580.98920, 480.31976],
        }
    )

    return con.create_table("xgb_table", df)


@pytest.fixture
def prediction_expr(feature_table, float_model_path):
    predict_fn = xo.expr.ml.make_quickgrove_udf(float_model_path)
    return feature_table.mutate(pred=predict_fn.on_expr)


@pytest.fixture
def mixed_prediction_expr(mixed_feature_table, mixed_model_path):
    predict_fn = xo.expr.ml.make_quickgrove_udf(mixed_model_path)
    return mixed_feature_table.mutate(pred=predict_fn.on_expr)


@pytest.fixture
def tmp_model_dir(tmpdir):
    # Create a temporary directory for the model
    model_dir = tmpdir.mkdir("models")
    return model_dir



import os

import pandas as pd
import snowflake.connector

import xorq as xo


def make_credential_defaults():
    return {
        "user": os.environ.get("SNOWFLAKE_USER"),
        "password": os.environ.get("SNOWFLAKE_PASSWORD"),
    }


def make_connection_defaults():
    return {
        "account": os.environ.get("SNOWFLAKE_ACCOUNT"),
        "role": os.environ.get("SNOWFLAKE_ROLE"),
        "warehouse": os.environ.get("SNOWFLAKE_WAREHOUSE"),
    }


def execute_statement(con, statement):
    (((resp,), *rest0), *rest1) = con.con.execute_string(statement)
    if rest0 or (resp != "Statement executed successfully."):
        raise ValueError


def make_connection(
    database,
    schema,
    **kwargs,
):
    con = xo.snowflake.connect(
        database=f"{database}/{schema}",
        **{
            **make_credential_defaults(),
            **make_connection_defaults(),
            **kwargs,
        },
    )
    return con


def make_raw_connection(database, schema, **kwargs):
    return snowflake.connector.connect(
        database=database,
        schema=schema,
        **{
            **make_credential_defaults(),
            **make_connection_defaults(),
            **kwargs,
        },
    )


def grant_create_database(con, role="public"):
    current_role = con.con.role
    statement = f"""
        USE ROLE accountadmin;
        GRANT CREATE DATABASE ON account TO ROLE {role};
        USE ROLE {current_role}
    """
    execute_statement(con, statement)


def grant_create_schema(con, role="public"):
    current_role = con.con.role
    statement = f"""
        USE ROLE accountadmin;
        GRANT CREATE SCHEMA ON account TO ROLE {role};
        USE ROLE {current_role}
    """
    execute_statement(con, statement)


def get_snowflake_last_modification_time(dt):
    (con, table, database, schema) = (
        dt.source,
        dt.name,
        dt.namespace.catalog,
        dt.namespace.database,
    )
    values = (
        con.table("TABLES", database=(database, "INFORMATION_SCHEMA"))[
            lambda t: t.TABLE_NAME == table
        ][lambda t: t.TABLE_SCHEMA == schema]
        .LAST_ALTERED.execute()
        .values
    )
    if not values:
        raise ValueError
    (value,) = values
    return value


def get_grants(con, role="public"):
    data = con.raw_sql(f"SHOW GRANTS TO ROLE {role}").fetchall()
    df = pd.DataFrame(
        data,
        columns=(
            "created_on",
            "privilege",
            "granted_on",
            "name",
            "granted_to",
            "grant_option",
            "granted_by_role_type",
            "granted_by",
        ),
    )
    return df


def get_session_query_df(con):
    stmt = """
    SELECT *
    FROM table(information_schema.query_history_by_session())
    ORDER BY start_time;
    """
    session_query_df = con.raw_sql(stmt).fetch_pandas_all()
    return session_query_df



import xorq.expr.relations as rel
import xorq.expr.udf as udf


opaque_ops = (
    rel.Read,
    rel.CachedNode,
    rel.RemoteTable,
    rel.FlightUDXF,
    rel.FlightExpr,
    udf.ExprScalarUDF,
)


def walk_nodes(node_types, expr):
    def process_node(op):
        match op:
            case rel.RemoteTable():
                if isinstance(op, node_types):
                    yield op
                yield from walk_nodes(
                    node_types,
                    op.remote_expr,
                )
            case rel.CachedNode():
                if isinstance(op, node_types):
                    yield op
                yield from walk_nodes(
                    node_types,
                    op.parent,
                )
            case rel.FlightExpr():
                if isinstance(op, node_types):
                    yield op
                yield from walk_nodes(node_types, op.input_expr)
            case rel.FlightUDXF():
                if isinstance(op, node_types):
                    yield op
                yield from walk_nodes(node_types, op.input_expr)
            case udf.ExprScalarUDF():
                if isinstance(op, node_types):
                    yield op
                yield from walk_nodes(
                    node_types,
                    op.computed_kwargs_expr,
                )
            case rel.Read():
                if isinstance(op, node_types):
                    yield op
            case _:
                if isinstance(op, opaque_ops):
                    raise ValueError(f"unhandled opaque op {type(op)}")
                yield from op.find(opaque_ops + tuple(node_types))

    def inner(rest, seen):
        if not rest:
            return seen
        op = rest.pop()
        seen.add(op)
        new = process_node(op)
        rest.update(set(new).difference(seen))
        return inner(rest, seen)

    initial_op = expr.op() if hasattr(expr, "op") else expr
    rest = process_node(initial_op)
    nodes = inner(set(rest), set())
    return tuple(node for node in nodes if isinstance(node, node_types))


def find_all_sources(expr):
    import xorq.vendor.ibis.expr.operations as ops

    node_types = (
        ops.DatabaseTable,
        ops.SQLQueryResult,
        rel.CachedNode,
        rel.Read,
        rel.RemoteTable,
        rel.FlightUDXF,
        rel.FlightExpr,
        # ExprScalarUDF has an expr we need to get to
        # FlightOperator has a dynamically generated connection: it should be passed a Profile instead
    )
    nodes = walk_nodes(node_types, expr)
    sources = tuple(
        source
        for (source, _) in set((node.source, node.source._profile) for node in nodes)
    )
    return sources



import pytest

import xorq as xo
from xorq.backends.let import Backend
from xorq.caching import (
    ParquetStorage,
)


@pytest.fixture
def cached_two(ls_con, batting, tmp_path):
    parquet_storage = ParquetStorage(source=ls_con, path=tmp_path)
    return (
        batting[lambda t: t.yearID > 2014]
        .cache()[lambda t: t.stint == 1]
        .cache(storage=parquet_storage)
    )


@pytest.fixture
def duck_batting_raw(batting_df):
    return xo.duckdb.connect().create_table(
        "batting_df",
        batting_df,
    )


@pytest.fixture
def duck_batting(duck_batting_raw):
    return duck_batting_raw


@pytest.fixture
def cached_two_joined(cached_two, duck_batting):
    return cached_two.join(
        duck_batting,
        duck_batting.columns,
    )


def test_ls_exists(batting):
    attr = getattr(batting, "ls", None)
    assert attr is not None


def test_ls_cache_nodes(cached_two):
    assert len(cached_two.ls.cached_nodes) == 2


def test_storage(cached_two):
    assert cached_two.ls.is_cached
    assert cached_two.ls.storage


def test_storages(cached_two):
    assert cached_two.ls.is_cached
    assert len(cached_two.ls.storages) == len(cached_two.ls.cached_nodes)


def test_backends(duck_batting_raw, cached_two, cached_two_joined):
    assert len(duck_batting_raw.ls.backends) == 1
    assert len(cached_two.ls.backends) == 2
    assert len(cached_two_joined.ls.backends) == 3


def test_is_multiengine(duck_batting_raw, cached_two, cached_two_joined):
    assert not duck_batting_raw.ls.is_multiengine
    assert cached_two.ls.is_multiengine
    assert cached_two_joined.ls.is_multiengine


def test_dts(cached_two, cached_two_joined):
    dts = cached_two.ls.dts
    assert len(dts) == 1
    assert not any(dt.source.name == Backend.name for dt in dts)

    dts = cached_two_joined.ls.dts
    assert len(dts) == 2
    assert not any(dt.source.name == Backend.name for dt in dts)


def test_is_cached(cached_two, cached_two_joined):
    assert cached_two.ls.is_cached
    assert not cached_two_joined.ls.is_cached


def test_has_cached(cached_two, cached_two_joined):
    els = cached_two.ls
    assert els.is_cached and els.has_cached
    els = cached_two_joined.ls
    assert not els.is_cached and els.has_cached


def test_uncached(cached_two):
    assert cached_two.ls.has_cached and not cached_two.ls.uncached.ls.has_cached


def test_uncached_one(cached_two):
    assert (
        cached_two.ls.is_cached
        and not cached_two.ls.uncached_one.ls.is_cached
        and cached_two.ls.uncached_one.ls.has_cached
    )


def test_exists(cached_two):
    storage = cached_two.ls.storage

    assert not cached_two.ls.exists()
    assert not tuple(storage.path.iterdir())

    xo.execute(cached_two)
    assert cached_two.ls.exists()
    assert len(tuple(storage.path.iterdir())) == 1

    (path,) = storage.path.iterdir()
    path.unlink()
    assert not cached_two.ls.exists()



import functools
import pickle

import pandas as pd
import pyarrow as pa

import xorq as xo
from xorq.common.utils.rbr_utils import (
    instrument_reader,
    streaming_split_exchange,
)
from xorq.flight import FlightServer
from xorq.flight.action import AddExchangeAction
from xorq.flight.exchanger import AbstractExchanger


SPLIT_KEY = "split"
MODEL_BINARY_KEY = "model_binary"


value = 0


def train_batch_df(df):
    global value
    value += len(df)
    return value


class IterativeSplitTrainExchanger(AbstractExchanger):
    @classmethod
    @property
    def exchange_f(cls):
        def train_batch(split_reader):
            df = split_reader.read_pandas()
            (split, *rest) = df[SPLIT_KEY].unique()
            assert not rest
            value = train_batch_df(df)
            batch = pa.RecordBatch.from_pydict(
                {
                    MODEL_BINARY_KEY: [pickle.dumps(value)],
                    SPLIT_KEY: [split],
                }
            )
            return batch

        return functools.partial(streaming_split_exchange, SPLIT_KEY, train_batch)

    @classmethod
    @property
    def schema_in_required(cls):
        return None

    @classmethod
    @property
    def schema_in_condition(cls):
        def condition(schema_in):
            return any(field.name == SPLIT_KEY for field in schema_in)

        return condition

    @classmethod
    @property
    def calc_schema_out(cls):
        def f(schema_in):
            split_field = next(field for field in schema_in if field.name == SPLIT_KEY)
            model_binary_field = pa.field(MODEL_BINARY_KEY, pa.binary())
            return pa.schema(
                (
                    model_binary_field,
                    split_field,
                )
            )

        return f

    @classmethod
    @property
    def description(cls):
        return "iteratively train model on data ordered by `split`"

    @classmethod
    @property
    def command(cls):
        return "iterative-split-train"


def train_test_split_union(expr, name=SPLIT_KEY, *args, **kwargs):
    splits = xo.expr.ml.train_test_splits(expr, *args, **kwargs)
    return xo.union(
        *(
            split.mutate(**{name: xo.literal(i, "int64")})
            for i, split in enumerate(splits)
        )
    )


con = xo.connect()
N = 10_000
df = pd.DataFrame({"a": range(N), "b": range(N, 2 * N)})
t = con.register(df, "t")
expr = train_test_split_union(
    t, unique_key="a", test_sizes=(0.2, 0.3, 0.5), random_seed=0
)


rbr_in = instrument_reader(xo.to_pyarrow_batches(expr), prefix="input ::")
with FlightServer() as server:
    client = server.client
    client.do_action(
        AddExchangeAction.name, IterativeSplitTrainExchanger, options=client._options
    )
    (fut, rbr_out) = client.do_exchange(IterativeSplitTrainExchanger.command, rbr_in)
    df_out = instrument_reader(rbr_out, prefix="output ::").read_pandas()
    print(fut.result())
    print(df_out.assign(model=df_out.model_binary.map(pickle.loads)))



from sklearn.feature_extraction.text import TfidfVectorizer

import xorq as xo
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.caching import ParquetStorage
from xorq.common.utils.defer_utils import deferred_read_parquet
from xorq.common.utils.import_utils import import_python
from xorq.expr.ml import (
    deferred_fit_transform_series_sklearn,
    train_test_splits,
)


m = import_python(xo.options.pins.get_path("hackernews_lib"))


deferred_fit_transform_tfidf = deferred_fit_transform_series_sklearn(
    col="title", cls=TfidfVectorizer, return_type=dt.Array(dt.float64)
)


con = xo.connect()
train_expr, test_expr = (
    deferred_read_parquet(
        con,
        xo.options.pins.get_path("hn-fetcher-input-small.parquet"),
        "fetcher-input",
    )
    .pipe(m.do_hackernews_fetcher_udxf)
    .pipe(
        train_test_splits,
        unique_key="id",
        test_sizes=(0.9, 0.1),
        random_seed=0,
    )
)


(deferred_model, model_udaf, deferred_transform) = deferred_fit_transform_tfidf(
    train_expr,
    storage=ParquetStorage(source=con),
)
bound_expr = test_expr.mutate(**{"transformed": deferred_transform.on_expr})


server, do_exchange = xo.expr.relations.flight_serve(bound_expr)
df = do_exchange(test_expr).read_pandas()
server.close()
try:
    do_exchange(test_expr.limit(10)).read_pandas()
    raise Exception("do_exchange should have raised")
except Exception:
    pass



from xorq.expr.relations import CachedNode, Read
from xorq.vendor.ibis import BaseBackend
from xorq.vendor.ibis.expr import operations as ops


def find_backend(op: ops.Node, use_default=False) -> tuple[BaseBackend, bool]:
    backends = set()
    has_unbound = False
    node_types = (
        ops.UnboundTable,
        ops.DatabaseTable,
        ops.SQLQueryResult,
        CachedNode,
        Read,
    )
    for table in op.find(node_types):
        if isinstance(table, ops.UnboundTable):
            has_unbound = True
        else:
            backends.add(table.source)

    if not backends and use_default:
        from xorq.config import _backend_init

        con = _backend_init()
        backends.add(con)

    return (
        backends.pop(),
        has_unbound,
    )  # TODO what happens if it has more than one backend



import pytest

import xorq as xo
from xorq.examples import Example
from xorq.examples.core import whitelist


def test_whitelist():
    assert [name in dir(xo.examples) for name in whitelist]


def test_attributes():
    for name in whitelist:
        example = getattr(xo.examples, name)
        assert isinstance(example, Example)

    with pytest.raises(AttributeError):
        xo.examples.missing



from __future__ import annotations

import datetime
import decimal
import functools
import operator
from typing import Any

import xorq.expr.datatypes as dt
import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.operations.temporal as tm
import xorq.vendor.ibis.expr.types as ir
from xorq.expr.relations import CachedNode, Read, RemoteTable, into_backend
from xorq.ibis_yaml.common import (
    TranslationContext,
    _translate_type,
    register_from_yaml_handler,
    translate_from_yaml,
    translate_to_yaml,
)

# ruff: noqa: F401
from xorq.ibis_yaml.udf import _scalar_udf_from_yaml, _scalar_udf_to_yaml
from xorq.ibis_yaml.utils import (
    freeze,
    load_storage_from_yaml,
    translate_storage,
)
from xorq.vendor.ibis.expr.operations.relations import Namespace


def should_register_node(node_dict):
    if "parent" in node_dict and isinstance(node_dict["parent"], dict):
        return True
    return False


@_translate_type.register(dt.Timestamp)
def _translate_timestamp_type(dtype: dt.Timestamp) -> dict:
    base = {"name": "Timestamp", "nullable": dtype.nullable}
    if dtype.timezone is not None:
        base["timezone"] = dtype.timezone
    return freeze(base)


@_translate_type.register(dt.Decimal)
def _translate_decimal_type(dtype: dt.Decimal) -> dict:
    base = {"name": "Decimal", "nullable": dtype.nullable}
    if dtype.precision is not None:
        base["precision"] = dtype.precision
    if dtype.scale is not None:
        base["scale"] = dtype.scale
    return freeze(base)


@_translate_type.register(dt.Array)
def _translate_array_type(dtype: dt.Array) -> dict:
    return freeze(
        {
            "name": "Array",
            "value_type": _translate_type(dtype.value_type),
            "nullable": dtype.nullable,
        }
    )


@_translate_type.register(dt.Map)
def _translate_map_type(dtype: dt.Map) -> dict:
    return freeze(
        {
            "name": "Map",
            "key_type": _translate_type(dtype.key_type),
            "value_type": _translate_type(dtype.value_type),
            "nullable": dtype.nullable,
        }
    )


@_translate_type.register(dt.Interval)
def _tranlate_type_interval(dtype: dt.Interval) -> dict:
    return freeze(
        {
            "name": "Interval",
            "unit": _translate_temporal_unit(dtype.unit),
            "nullable": dtype.nullable,
        }
    )


@_translate_type.register(dt.Struct)
def _translate_struct_type(dtype: dt.Struct) -> dict:
    return freeze(
        {
            "name": "Struct",
            "fields": {
                name: _translate_type(field_type)
                for name, field_type in zip(dtype.names, dtype.types)
            },
            "nullable": dtype.nullable,
        }
    )


def _translate_temporal_unit(unit: tm.IntervalUnit) -> dict:
    if unit.is_date():
        unit_name = "DateUnit"
    elif unit.is_time():
        unit_name = "TimeUnit"
    else:
        unit_name = "IntervalUnit"
    return freeze({"name": unit_name, "value": unit.value})


def _translate_literal_value(value: Any, dtype: dt.DataType) -> Any:
    if value is None:
        return None
    elif isinstance(value, (bool, int, float, str)):
        return value
    elif isinstance(value, decimal.Decimal):
        return str(value)
    elif isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
        return value.isoformat()
    elif isinstance(value, list):
        return [_translate_literal_value(v, dtype.value_type) for v in value]
    elif isinstance(value, dict):
        return {
            _translate_literal_value(k, dtype.key_type): _translate_literal_value(
                v, dtype.value_type
            )
            for k, v in value.items()
        }
    else:
        return value


@translate_to_yaml.register(ir.Expr)
def _expr_to_yaml(expr: ir.Expr, context: any) -> dict:
    return translate_to_yaml(expr.op(), context)


@translate_to_yaml.register(ops.WindowFunction)
def _window_function_to_yaml(
    op: ops.WindowFunction, context: TranslationContext
) -> dict:
    result = {
        "op": "WindowFunction",
        "args": [translate_to_yaml(op.func, context)],
        "type": _translate_type(op.dtype),
    }

    if op.group_by:
        result["group_by"] = [translate_to_yaml(expr, context) for expr in op.group_by]

    if op.order_by:
        result["order_by"] = [translate_to_yaml(expr, context) for expr in op.order_by]

    if op.start is not None:
        result["start"] = (
            translate_to_yaml(op.start.value, context)["value"]
            if isinstance(op.start, ops.WindowBoundary)
            else op.start
        )

    if op.end is not None:
        result["end"] = (
            translate_to_yaml(op.end.value, context)["value"]
            if isinstance(op.end, ops.WindowBoundary)
            else op.end
        )

    return freeze(result)


@register_from_yaml_handler("WindowFunction")
def _window_function_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    func = translate_from_yaml(yaml_dict["args"][0], context)
    group_by = [translate_from_yaml(g, context) for g in yaml_dict.get("group_by", [])]
    order_by = [translate_from_yaml(o, context) for o in yaml_dict.get("order_by", [])]
    start = ibis.literal(yaml_dict["start"]) if "start" in yaml_dict else None
    end = ibis.literal(yaml_dict["end"]) if "end" in yaml_dict else None
    window = ibis.window(
        group_by=group_by, order_by=order_by, preceding=start, following=end
    )
    return func.over(window)


@translate_to_yaml.register(ops.WindowBoundary)
def _window_boundary_to_yaml(
    op: ops.WindowBoundary, context: TranslationContext
) -> dict:
    return freeze(
        {
            "op": "WindowBoundary",
            "value": translate_to_yaml(op.value, context),
            "preceding": op.preceding,
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("WindowBoundary")
def _window_boundary_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    value = translate_from_yaml(yaml_dict["value"], context)
    return ops.WindowBoundary(value, preceding=yaml_dict["preceding"])


@translate_to_yaml.register(ops.Node)
def _base_op_to_yaml(op: ops.Node, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
            "args": [
                translate_to_yaml(arg, context)
                for arg in op.args
                if isinstance(arg, (ops.Value, ops.Node))
            ],
        }
    )


@translate_to_yaml.register(ops.UnboundTable)
def _unbound_table_to_yaml(op: ops.UnboundTable, context: TranslationContext) -> dict:
    schema_id = context.schema_registry.register_schema(op.schema)
    namespace_dict = freeze(
        {
            "catalog": op.namespace.catalog,
            "database": op.namespace.database,
        }
    )
    return freeze(
        {
            "op": "UnboundTable",
            "name": op.name,
            "schema_ref": schema_id,
            "namespace": namespace_dict,
        }
    )


@register_from_yaml_handler("UnboundTable")
def _unbound_table_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    table_name = yaml_dict["name"]

    schema_ref = yaml_dict["schema_ref"]
    try:
        schema_def = context.definitions["schemas"][schema_ref]
    except KeyError:
        raise ValueError(f"Schema {schema_ref} not found in definitions")
    namespace_dict = yaml_dict.get("namespace", {})
    catalog = namespace_dict.get("catalog")
    database = namespace_dict.get("database")
    schema = {
        name: _type_from_yaml(dtype_yaml) for name, dtype_yaml in schema_def.items()
    }
    # TODO: use UnboundTable node to construct instead of builder API
    return ibis.table(schema, name=table_name, catalog=catalog, database=database)


@translate_to_yaml.register(ops.DatabaseTable)
def _database_table_to_yaml(op: ops.DatabaseTable, context: TranslationContext) -> dict:
    profile_name = op.source._profile.hash_name
    schema_id = context.schema_registry.register_schema(op.schema)
    namespace_dict = freeze(
        {
            "catalog": op.namespace.catalog,
            "database": op.namespace.database,
        }
    )

    node_dict = freeze(
        {
            "op": "DatabaseTable",
            "table": op.name,
            "schema_ref": schema_id,
            "profile": profile_name,
            "namespace": namespace_dict,
        }
    )

    if should_register_node(node_dict) and hasattr(
        context.schema_registry, "register_node"
    ):
        node_hash = context.schema_registry.register_node(node_dict)
        return freeze({"node_ref": node_hash})

    return node_dict


@register_from_yaml_handler("DatabaseTable")
def database_table_from_yaml(yaml_dict: dict, context: TranslationContext) -> ibis.Expr:
    profile_name = yaml_dict.get("profile")
    table_name = yaml_dict.get("table")
    namespace_dict = yaml_dict.get("namespace", {})
    catalog = namespace_dict.get("catalog")
    database = namespace_dict.get("database")
    # we should validate that schema is the same
    schema_ref = yaml_dict.get("schema_ref")
    schema_def = context.definitions["schemas"][schema_ref]
    fields = []
    for name, dtype_yaml in schema_def.items():
        dtype = _type_from_yaml(dtype_yaml)
        fields.append((name, dtype))
    schema = ibis.Schema.from_tuples(fields)

    try:
        con = context.profiles[profile_name]
    except KeyError:
        raise ValueError(f"Profile {profile_name!r} not found in context.profiles")
    return ops.DatabaseTable(
        schema=schema,
        source=con,
        name=table_name,
        namespace=Namespace(catalog=catalog, database=database),
    ).to_expr()


@translate_to_yaml.register(CachedNode)
def _cached_node_to_yaml(op: CachedNode, context: any) -> dict:
    schema_id = context.schema_registry.register_schema(op.schema)
    # source should be called profile_name

    return freeze(
        {
            "op": "CachedNode",
            "schema_ref": schema_id,
            "parent": translate_to_yaml(op.parent, context),
            "source": op.source._profile.hash_name,
            "storage": translate_storage(op.storage, context),
        }
    )


@register_from_yaml_handler("CachedNode")
def _cached_node_from_yaml(yaml_dict: dict, context: any) -> ibis.Expr:
    schema_ref = yaml_dict["schema_ref"]
    try:
        schema_def = context.definitions["schemas"][schema_ref]
    except KeyError:
        raise ValueError(f"Schema {schema_ref} not found in definitions")

    schema = {
        name: _type_from_yaml(dtype_yaml) for name, dtype_yaml in schema_def.items()
    }

    parent_expr = translate_from_yaml(yaml_dict["parent"], context)
    profile_name = yaml_dict.get("source")
    try:
        source = context.profiles[profile_name]
    except KeyError:
        raise ValueError(f"Profile {profile_name!r} not found in context.profiles")
    storage = load_storage_from_yaml(yaml_dict["storage"], context)

    op = CachedNode(
        schema=schema,
        parent=parent_expr,
        source=source,
        storage=storage,
    )
    return op.to_expr()


@translate_to_yaml.register(RemoteTable)
def _remotetable_to_yaml(op: RemoteTable, context: any) -> dict:
    profile_name = op.source._profile.hash_name
    remote_expr_yaml = translate_to_yaml(op.remote_expr, context)
    schema_id = context.schema_registry.register_schema(op.schema)
    # TODO: change profile to profile_name
    return freeze(
        {
            "op": "RemoteTable",
            "table": op.name,
            "schema_ref": schema_id,
            "profile": profile_name,
            "remote_expr": remote_expr_yaml,
        }
    )


@register_from_yaml_handler("RemoteTable")
def _remotetable_from_yaml(yaml_dict: dict, context: any) -> ibis.Expr:
    profile_name = yaml_dict.get("profile")
    table_name = yaml_dict.get("table")
    remote_expr_yaml = yaml_dict.get("remote_expr")
    if profile_name is None:
        raise ValueError(
            "Missing keys in RemoteTable YAML; ensure 'profile_name' are present."
        )
    try:
        con = context.profiles[profile_name]
    except KeyError:
        raise ValueError(f"Profile {profile_name!r} not found in context.profiles")

    remote_expr = translate_from_yaml(remote_expr_yaml, context)

    remote_table_expr = into_backend(remote_expr, con, table_name)
    return remote_table_expr


@translate_to_yaml.register(Read)
def _read_to_yaml(op: Read, context: TranslationContext) -> dict:
    schema_id = context.schema_registry.register_schema(op.schema)
    profile_hash_name = (
        op.source._profile.hash_name if hasattr(op.source, "_profile") else None
    )

    return freeze(
        {
            "op": "Read",
            "method_name": op.method_name,
            "name": op.name,
            "schema_ref": schema_id,
            "profile": profile_hash_name,
            "read_kwargs": freeze(op.read_kwargs if op.read_kwargs else {}),
        }
    )


@register_from_yaml_handler("Read")
def _read_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    schema_ref = yaml_dict["schema_ref"]
    schema_def = context.definitions["schemas"][schema_ref]

    schema = {
        name: _type_from_yaml(dtype_yaml) for name, dtype_yaml in schema_def.items()
    }

    profile_hash_name = yaml_dict.get("profile")

    source = context.profiles[profile_hash_name]

    read_op = Read(
        method_name=yaml_dict["method_name"],
        name=yaml_dict["name"],
        schema=schema,
        source=source,
        read_kwargs=yaml_dict.get("read_kwargs", {}),
    )

    return read_op.to_expr()


@translate_to_yaml.register(ops.Literal)
def _literal_to_yaml(op: ops.Literal, context: TranslationContext) -> dict:
    value = _translate_literal_value(op.value, op.dtype)
    return freeze({"op": "Literal", "value": value, "type": _translate_type(op.dtype)})


@register_from_yaml_handler("Literal")
def _literal_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    value = yaml_dict["value"]
    dtype = _type_from_yaml(yaml_dict["type"])
    return ibis.literal(value, type=dtype)


@translate_to_yaml.register(ops.ValueOp)
def _value_op_to_yaml(op: ops.ValueOp, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
            "type": _translate_type(op.dtype),
            "args": [
                translate_to_yaml(arg, context)
                for arg in op.args
                if isinstance(arg, (ops.Value, ops.Node))
            ],
        }
    )


@register_from_yaml_handler("ValueOp")
def _value_op_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    method_name = yaml_dict["op"].lower()
    method = getattr(args[0], method_name)
    return method(*args[1:])


@translate_to_yaml.register(ops.Lag)
def _lag_to_yaml(op: ops.Lag, context: TranslationContext) -> dict:
    result = {
        "op": "Lag",
        "arg": translate_to_yaml(op.arg, context),
        "type": _translate_type(op.dtype),
    }

    if op.offset is not None:
        result["offset"] = translate_to_yaml(op.offset, context)

    if op.default is not None:
        result["default"] = translate_to_yaml(op.default, context)

    node_dict = freeze(result)

    return node_dict


@register_from_yaml_handler("Lag")
def _lag_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["arg"], context)

    offset = None
    if "offset" in yaml_dict:
        offset = translate_from_yaml(yaml_dict["offset"], context)

    default = None
    if "default" in yaml_dict:
        default = translate_from_yaml(yaml_dict["default"], context)

    return arg.lag(offset, default)


@translate_to_yaml.register(ops.StringUnary)
def _string_unary_to_yaml(op: ops.StringUnary, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
            "args": [translate_to_yaml(op.arg, context)],
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("StringUnary")
def _string_unary_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    method_name = yaml_dict["op"].lower()
    return getattr(arg, method_name)()


@translate_to_yaml.register(ops.Substring)
def _substring_to_yaml(op: ops.Substring, context: TranslationContext) -> dict:
    args = [
        translate_to_yaml(op.arg, context),
        translate_to_yaml(op.start, context),
    ]
    if op.length is not None:
        args.append(translate_to_yaml(op.length, context))
    return freeze({"op": "Substring", "args": args, "type": _translate_type(op.dtype)})


@register_from_yaml_handler("Substring")
def _substring_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return args[0].substr(args[1], args[2] if len(args) > 2 else None)


@translate_to_yaml.register(ops.StringLength)
def _string_length_to_yaml(op: ops.StringLength, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "StringLength",
            "args": [translate_to_yaml(op.arg, context)],
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("StringLength")
def _string_length_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return arg.length()


@translate_to_yaml.register(ops.StringConcat)
def _string_concat_to_yaml(op: ops.StringConcat, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "StringConcat",
            "args": [translate_to_yaml(arg, context) for arg in op.arg],
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("StringConcat")
def _string_concat_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return functools.reduce(lambda x, y: x.concat(y), args)


@register_from_yaml_handler("StringContains")
def _string_contains_from_yaml(yaml_dict: dict, context: TranslationContext) -> dict:
    haystack, needle = (translate_from_yaml(arg, context) for arg in yaml_dict["args"])
    return ops.StringContains(haystack, needle).to_expr()


@translate_to_yaml.register(ops.BinaryOp)
def _binary_op_to_yaml(op: ops.BinaryOp, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
            "args": [
                translate_to_yaml(op.left, context),
                translate_to_yaml(op.right, context),
            ],
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("BinaryOp")
def _binary_op_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    op_name = yaml_dict["op"].lower()
    return getattr(args[0], op_name)(args[1])


@translate_to_yaml.register(ops.Filter)
def _filter_to_yaml(op: ops.Filter, context: TranslationContext) -> dict:
    node_dict = freeze(
        {
            "op": "Filter",
            "parent": translate_to_yaml(op.parent, context),
            "predicates": [translate_to_yaml(pred, context) for pred in op.predicates],
        }
    )

    if should_register_node(node_dict) and hasattr(
        context.schema_registry, "register_node"
    ):
        node_hash = context.schema_registry.register_node(node_dict)
        return freeze({"node_ref": node_hash})

    return node_dict


@register_from_yaml_handler("Filter")
def _filter_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = translate_from_yaml(yaml_dict["parent"], context)
    predicates = [
        translate_from_yaml(pred, context) for pred in yaml_dict["predicates"]
    ]
    filter_op = ops.Filter(parent, predicates)
    return filter_op.to_expr()


@translate_to_yaml.register(ops.Project)
def _project_to_yaml(op: ops.Project, context: TranslationContext) -> dict:
    node_dict = {
        "op": "Project",
        "parent": translate_to_yaml(op.parent, context),
        "values": {
            name: translate_to_yaml(val, context) for name, val in op.values.items()
        },
    }

    if should_register_node(node_dict):
        node_hash = context.schema_registry.register_node(freeze(node_dict))
        return freeze({"node_ref": node_hash})

    return freeze(node_dict)


@register_from_yaml_handler("Project")
def _project_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = translate_from_yaml(yaml_dict["parent"], context)

    values_dict = yaml_dict.get("values", {})

    values = {
        name: translate_from_yaml(val, context) for name, val in values_dict.items()
    }

    projected = parent.projection(values)
    return projected


@translate_to_yaml.register(ops.Aggregate)
def _aggregate_to_yaml(op: ops.Aggregate, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Aggregate",
            "parent": translate_to_yaml(op.parent, context),
            "by": [translate_to_yaml(group, context) for group in op.groups.values()],
            "metrics": {
                name: translate_to_yaml(metric, context)
                for name, metric in op.metrics.items()
            },
        }
    )


@register_from_yaml_handler("Aggregate")
def _aggregate_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = translate_from_yaml(yaml_dict["parent"], context)
    groups = tuple(
        translate_from_yaml(group, context) for group in yaml_dict.get("by", [])
    )

    metrics = {
        name: translate_from_yaml(metric, context)
        for name, metric in yaml_dict.get("metrics", {}).items()
    }

    result = (
        parent.group_by(list(groups)).aggregate(metrics)
        if groups
        else parent.aggregate(metrics)
    )
    return result


@translate_to_yaml.register(ops.JoinChain)
def _join_to_yaml(op: ops.JoinChain, context: TranslationContext) -> dict:
    node_dict = {
        "op": "JoinChain",
        "first": translate_to_yaml(op.first, context),
        "rest": [
            {
                "how": link.how,
                "table": translate_to_yaml(link.table, context),
                "predicates": [
                    translate_to_yaml(pred, context) for pred in link.predicates
                ],
            }
            for link in op.rest
        ],
    }

    node_dict["values"] = {
        name: translate_to_yaml(val, context) for name, val in op.values.items()
    }

    if should_register_node(node_dict):
        node_hash = context.schema_registry.register_node(freeze(node_dict))
        return freeze({"node_ref": node_hash})
    return freeze(node_dict)


@register_from_yaml_handler("JoinChain")
def _join_chain_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    first = translate_from_yaml(yaml_dict["first"], context)
    result = first

    for join in yaml_dict["rest"]:
        table = translate_from_yaml(join["table"], context)
        predicates = [translate_from_yaml(pred, context) for pred in join["predicates"]]
        result = result.join(table, predicates, how=join["how"])

    values = {
        name: translate_from_yaml(val, context)
        for name, val in yaml_dict["values"].items()
    }
    result = result.select(values)
    return result


@translate_to_yaml.register(ops.Sort)
def _sort_to_yaml(op: ops.Sort, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Sort",
            "parent": translate_to_yaml(op.parent, context),
            "keys": [translate_to_yaml(key, context) for key in op.keys],
        }
    )


@register_from_yaml_handler("Sort")
def _sort_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = translate_from_yaml(yaml_dict["parent"], context)
    keys = tuple(translate_from_yaml(key, context) for key in yaml_dict["keys"])
    sort_op = ops.Sort(parent, keys=keys)
    return sort_op.to_expr()


@translate_to_yaml.register(ops.SortKey)
def _sort_key_to_yaml(op: ops.SortKey, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "SortKey",
            "arg": translate_to_yaml(op.expr, context),
            "ascending": op.ascending,
            "nulls_first": op.nulls_first,
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("SortKey")
def _sort_key_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    expr = translate_from_yaml(yaml_dict["arg"], context)
    ascending = yaml_dict.get("ascending", True)
    nulls_first = yaml_dict.get("nulls_first", False)
    return ops.SortKey(expr, ascending=ascending, nulls_first=nulls_first).to_expr()


@translate_to_yaml.register(ops.Limit)
def _limit_to_yaml(op: ops.Limit, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Limit",
            "parent": translate_to_yaml(op.parent, context),
            "n": op.n,
            "offset": op.offset,
        }
    )


@register_from_yaml_handler("Limit")
def _limit_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = translate_from_yaml(yaml_dict["parent"], context)
    return parent.limit(yaml_dict["n"], offset=yaml_dict["offset"])


@translate_to_yaml.register(ops.ScalarSubquery)
def _scalar_subquery_to_yaml(
    op: ops.ScalarSubquery, context: TranslationContext
) -> dict:
    return freeze(
        {
            "op": "ScalarSubquery",
            "args": [translate_to_yaml(arg, context) for arg in op.args],
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("ScalarSubquery")
def _scalar_subquery_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    subquery = translate_from_yaml(yaml_dict["args"][0], context)
    return ops.ScalarSubquery(subquery).to_expr()


@translate_to_yaml.register(ops.ExistsSubquery)
def _exists_subquery_to_yaml(
    op: ops.ExistsSubquery, context: TranslationContext
) -> dict:
    return freeze(
        {
            "op": "ExistsSubquery",
            "rel": translate_to_yaml(op.rel, context),
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("ExistsSubquery")
def _exists_subquery_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    rel = translate_from_yaml(yaml_dict["rel"], context)
    return ops.ExistsSubquery(rel).to_expr()


@translate_to_yaml.register(ops.InSubquery)
def _in_subquery_to_yaml(op: ops.InSubquery, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "InSubquery",
            "needle": translate_to_yaml(op.needle, context),
            "haystack": translate_to_yaml(op.rel, context),
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("InSubquery")
def _in_subquery_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    needle = translate_from_yaml(yaml_dict["needle"], context)
    haystack = translate_from_yaml(yaml_dict["haystack"], context)
    return ops.InSubquery(haystack, needle).to_expr()


@translate_to_yaml.register(ops.Field)
def _field_to_yaml(op: ops.Field, context: TranslationContext) -> dict:
    result = {
        "op": "Field",
        "name": op.name,
        "relation": translate_to_yaml(op.rel, context),
        "type": _translate_type(op.dtype),
    }

    if op.args and len(op.args) >= 2 and isinstance(op.args[1], str):
        underlying_name = op.args[1]
        if underlying_name != op.name:
            result["original_name"] = underlying_name

    node_dict = freeze(result)

    if hasattr(context.schema_registry, "register_node"):
        node_hash = context.schema_registry.register_node(node_dict)
        return freeze({"node_ref": node_hash})

    return node_dict


@register_from_yaml_handler("Field")
def field_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    relation = translate_from_yaml(yaml_dict["relation"], context)
    target_name = yaml_dict["name"]
    source_name = yaml_dict.get("original_name", target_name)
    field = relation[source_name]
    if target_name != source_name:
        field = field.name(target_name)

    return freeze(field)


@translate_to_yaml.register(ops.InValues)
def _in_values_to_yaml(op: ops.InValues, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "InValues",
            "args": [
                translate_to_yaml(op.value, context),
                *[translate_to_yaml(opt, context) for opt in op.options],
            ],
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("InValues")
def _in_values_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    value = translate_from_yaml(yaml_dict["args"][0], context)
    options = tuple(translate_from_yaml(opt, context) for opt in yaml_dict["args"][1:])
    return ops.InValues(value, options).to_expr()


@translate_to_yaml.register(ops.SimpleCase)
def _simple_case_to_yaml(op: ops.SimpleCase, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "SimpleCase",
            "base": translate_to_yaml(op.base, context),
            "cases": [translate_to_yaml(case, context) for case in op.cases],
            "results": [translate_to_yaml(result, context) for result in op.results],
            "default": translate_to_yaml(op.default, context),
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("SimpleCase")
def _simple_case_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    base = translate_from_yaml(yaml_dict["base"], context)
    cases = tuple(translate_from_yaml(case, context) for case in yaml_dict["cases"])
    results = tuple(
        translate_from_yaml(result, context) for result in yaml_dict["results"]
    )
    default = translate_from_yaml(yaml_dict["default"], context)
    return ops.SimpleCase(base, cases, results, default).to_expr()


@translate_to_yaml.register(ops.IfElse)
def _if_else_to_yaml(op: ops.IfElse, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "IfElse",
            "bool_expr": translate_to_yaml(op.bool_expr, context),
            "true_expr": translate_to_yaml(op.true_expr, context),
            "false_null_expr": translate_to_yaml(op.false_null_expr, context),
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("IfElse")
def _if_else_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    bool_expr = translate_from_yaml(yaml_dict["bool_expr"], context)
    true_expr = translate_from_yaml(yaml_dict["true_expr"], context)
    false_null_expr = translate_from_yaml(yaml_dict["false_null_expr"], context)
    return ops.IfElse(bool_expr, true_expr, false_null_expr).to_expr()


@translate_to_yaml.register(ops.Coalesce)
def _coalesce_to_yaml(op: ops.Coalesce, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Coalesce",
            "args": [translate_to_yaml(arg, context) for arg in op.arg],
            "type": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("Coalesce")
def _coalesce_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return ibis.coalesce(*args)


@translate_to_yaml.register(ops.CountDistinct)
def _count_distinct_to_yaml(op: ops.CountDistinct, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "CountDistinct",
            "args": [translate_to_yaml(op.arg, context)],
            "type": _translate_type(op.dtype),
        }
    )


@translate_to_yaml.register(ops.RankBase)
def _rank_base_to_yaml(op: ops.RankBase, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": type(op).__name__,
        }
    )


@register_from_yaml_handler("MinRank")
def _min_rank_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    return ibis.rank()


@register_from_yaml_handler("RowNumber")
def _row_number_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    return ibis.row_number()


@register_from_yaml_handler("CountDistinct")
def _count_distinct_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return arg.nunique()


@translate_to_yaml.register(ops.SelfReference)
def _self_reference_to_yaml(op: ops.SelfReference, context: TranslationContext) -> dict:
    result = {"op": "SelfReference", "identifier": op.identifier}
    if op.args:
        result["args"] = [translate_to_yaml(op.args[0], context)]
    return freeze(result)


@register_from_yaml_handler("SelfReference")
def _self_reference_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    if "args" in yaml_dict and yaml_dict["args"]:
        underlying = translate_from_yaml(yaml_dict["args"][0], context)
    else:
        if underlying is None:
            raise NotImplementedError("No relation available for SelfReference")

    identifier = yaml_dict.get("identifier", 0)
    ref = ops.SelfReference(underlying, identifier=identifier)

    return ref.to_expr()


@translate_to_yaml.register(ops.DropColumns)
def _drop_columns_to_yaml(op: ops.DropColumns, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "DropColumns",
            "parent": translate_to_yaml(op.parent, context),
            "columns_to_drop": list(op.columns_to_drop),
        }
    )


@register_from_yaml_handler("DropColumns")
def _drop_columns_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    parent = translate_from_yaml(yaml_dict["parent"], context)
    columns = frozenset(yaml_dict["columns_to_drop"])
    op = ops.DropColumns(parent, columns)
    return op.to_expr()


@translate_to_yaml.register(ops.SearchedCase)
def _searched_case_to_yaml(op: ops.SearchedCase, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "SearchedCase",
            "cases": [translate_to_yaml(case, context) for case in op.cases],
            "results": [translate_to_yaml(result, context) for result in op.results],
            "default": translate_to_yaml(op.default, context),
            "dtype": _translate_type(op.dtype),
        }
    )


@register_from_yaml_handler("SearchedCase")
def _searched_case_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    cases = [translate_from_yaml(case, context) for case in yaml_dict["cases"]]
    results = [translate_from_yaml(result, context) for result in yaml_dict["results"]]
    default = translate_from_yaml(yaml_dict["default"], context)
    op = ops.SearchedCase(cases, results, default)
    return op.to_expr()


@register_from_yaml_handler("View")
def _view_from_yaml(yaml_dict: dict, context: any) -> ir.Expr:
    underlying = translate_from_yaml(yaml_dict["args"][0], context)
    alias = yaml_dict.get("name")
    if alias:
        return underlying.alias(alias)
    return underlying


@register_from_yaml_handler("Mean")
def _mean_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return args[0].mean()


@register_from_yaml_handler("Add", "Subtract", "Multiply", "Divide")
def _binary_arithmetic_from_yaml(
    yaml_dict: dict, context: TranslationContext
) -> ir.Expr:
    left = translate_from_yaml(yaml_dict["args"][0], context)
    right = translate_from_yaml(yaml_dict["args"][1], context)
    op_map = {
        "Add": operator.add,
        "Subtract": operator.sub,
        "Multiply": operator.mul,
        "Divide": operator.truediv,
    }
    op_func = op_map.get(yaml_dict["op"])
    if op_func is None:
        raise ValueError(f"Unsupported arithmetic operation: {yaml_dict['op']}")
    return op_func(left, right)


@register_from_yaml_handler("Repeat")
def _repeat_from_yaml(
    yaml_dict: dict, context: TranslationContext
) -> ibis.expr.types.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    times = translate_from_yaml(yaml_dict["args"][1], context)
    return ops.Repeat(arg, times).to_expr()


@register_from_yaml_handler("Sum")
def _sum_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return args[0].sum()


@register_from_yaml_handler("Min")
def _min_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return args[0].min()


@register_from_yaml_handler("Max")
def _max_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return args[0].max()


@register_from_yaml_handler("Abs")
def _abs_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return arg.abs()


@register_from_yaml_handler("Modulus")
def _sum_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return args[0].sum()


@register_from_yaml_handler("Count")
def _count_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return arg.count()


@register_from_yaml_handler("JoinReference")
def _join_reference_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    table_yaml = yaml_dict["args"][0]
    return translate_from_yaml(table_yaml, context)


@register_from_yaml_handler(
    "Equals", "NotEquals", "GreaterThan", "GreaterEqual", "LessThan", "LessEqual"
)
def _binary_compare_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    left = translate_from_yaml(yaml_dict["args"][0], context)
    right = translate_from_yaml(yaml_dict["args"][1], context)

    op_map = {
        "Equals": operator.eq,
        "NotEquals": operator.ne,
        "GreaterThan": operator.gt,
        "GreaterEqual": operator.ge,
        "LessThan": operator.lt,
        "LessEqual": operator.le,
    }

    op_func = op_map.get(yaml_dict["op"])
    if op_func is None:
        raise ValueError(f"Unsupported comparison operation: {yaml_dict['op']}")
    return op_func(left, right)


@register_from_yaml_handler("Between")
def _between_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    return args[0].between(args[1], args[2])


@register_from_yaml_handler("Greater", "Less")
def _boolean_ops_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict["args"]]
    op_name = yaml_dict["op"]
    op_map = {
        "Greater": operator.gt,
        "Less": operator.lt,
    }
    return op_map[op_name](*args)


@register_from_yaml_handler("And")
def _boolean_and_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict.get("args", [])]
    if not args:
        raise ValueError("And operator requires at least one argument")
    return functools.reduce(operator.and_, args)


@register_from_yaml_handler("Or")
def _boolean_or_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = [translate_from_yaml(arg, context) for arg in yaml_dict.get("args", [])]
    if not args:
        raise ValueError("Or operator requires at least one argument")
    return functools.reduce(operator.or_, args)


@register_from_yaml_handler("Not")
def _not_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return ~arg


@register_from_yaml_handler("IsNull")
def _is_null_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return arg.isnull()


@register_from_yaml_handler("NotNull")
def _not_null_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return arg.notnull()


@register_from_yaml_handler(
    "ExtractYear",
    "ExtractMonth",
    "ExtractDay",
    "ExtractHour",
    "ExtractMinute",
    "ExtractSecond",
)
def _extract_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    op_map = {
        "ExtractYear": operator.methodcaller("year"),
        "ExtractMonth": operator.methodcaller("month"),
        "ExtractDay": operator.methodcaller("day"),
        "ExtractHour": operator.methodcaller("hour"),
        "ExtractMinute": operator.methodcaller("minute"),
        "ExtractSecond": operator.methodcaller("second"),
    }
    return op_map[yaml_dict["op"]](arg)


@register_from_yaml_handler("TimestampDiff")
def _timestamp_diff_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    left = translate_from_yaml(yaml_dict["args"][0], context)
    right = translate_from_yaml(yaml_dict["args"][1], context)
    return left - right


@register_from_yaml_handler("TimestampAdd", "TimestampSub")
def _timestamp_arithmetic_from_yaml(
    yaml_dict: dict, context: TranslationContext
) -> ir.Expr:
    timestamp = translate_from_yaml(yaml_dict["args"][0], context)
    interval = translate_from_yaml(yaml_dict["args"][1], context)
    if yaml_dict["op"] == "TimestampAdd":
        return timestamp + interval
    else:
        return timestamp - interval


@register_from_yaml_handler("Cast")
def _cast_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    target_dtype = _type_from_yaml(yaml_dict["type"])
    return arg.cast(target_dtype)


@register_from_yaml_handler("Hash")
def _hash_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)
    return ops.Hash(arg).to_expr()


@register_from_yaml_handler("CountStar")
def _count_star_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    arg = translate_from_yaml(yaml_dict["args"][0], context)

    return ops.CountStar(arg).to_expr()


@register_from_yaml_handler("StringSQLLike")
def _string_sql_like_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = yaml_dict.get("args", [])
    if not args:
        raise ValueError("Missing arguments for StringSQLLike operator")

    col = translate_from_yaml(args[0], context)

    if len(args) >= 2:
        pattern_expr = translate_from_yaml(args[1], context)
    else:
        pattern_value = args[0].get("value")
        if pattern_value is None:
            pattern_value = yaml_dict.get("value")
        if pattern_value is None:
            raise ValueError("Missing pattern for StringSQLLike operator")
        pattern_expr = ibis.literal(pattern_value, type=dt.String())

    escape = yaml_dict.get("escape")

    return ops.StringSQLLike(col, pattern_expr, escape=escape).to_expr()


@translate_to_yaml.register(ops.StringJoin)
def _string_join_to_yaml(op: ops.StringJoin, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "StringJoin",
            "args": [translate_to_yaml(arg, context) for arg in op.arg],
            "sep": translate_to_yaml(op.sep, context),
        }
    )


@register_from_yaml_handler("StringJoin")
def _string_join_from_yaml(yaml_dict: dict, context: TranslationContext) -> ir.Expr:
    args = tuple(translate_from_yaml(arg, context) for arg in yaml_dict["args"])
    sep = translate_from_yaml(yaml_dict["sep"], context)

    return ops.StringJoin(args, sep).to_expr()


def _type_from_yaml(yaml_dict: dict) -> dt.DataType:
    type_name = yaml_dict["name"]
    base_type = REVERSE_TYPE_REGISTRY.get(type_name)
    if base_type is None:
        raise ValueError(f"Unknown type: {type_name}")
    if callable(base_type) and not isinstance(base_type, dt.DataType):
        base_type = base_type(yaml_dict)
    elif (
        "nullable" in yaml_dict
        and isinstance(base_type, dt.DataType)
        and not isinstance(base_type, (tm.IntervalUnit, dt.Timestamp))
    ):
        base_type = base_type.copy(nullable=yaml_dict["nullable"])
    return base_type


REVERSE_TYPE_REGISTRY = {
    "Int8": dt.Int8(),
    "Int16": dt.Int16(),
    "Int32": dt.Int32(),
    "Int64": dt.Int64(),
    "UInt8": dt.UInt8(),
    "UInt16": dt.UInt16(),
    "UInt32": dt.UInt32(),
    "UInt64": dt.UInt64(),
    "Float32": dt.Float32(),
    "Float64": dt.Float64(),
    "String": dt.String(),
    "Boolean": dt.Boolean(),
    "Date": dt.Date(),
    "Time": dt.Time(),
    "Binary": dt.Binary(),
    "JSON": dt.JSON(),
    "Null": dt.null,
    "Timestamp": lambda yaml_dict: dt.Timestamp(
        nullable=yaml_dict.get("nullable", True)
    ),
    "Decimal": lambda yaml_dict: dt.Decimal(
        precision=yaml_dict.get("precision"),
        scale=yaml_dict.get("scale"),
        nullable=yaml_dict.get("nullable", True),
    ),
    "IntervalUnit": lambda yaml_dict: tm.IntervalUnit(
        yaml_dict["value"] if isinstance(yaml_dict, dict) else yaml_dict
    ),
    "Interval": lambda yaml_dict: dt.Interval(
        unit=_type_from_yaml(yaml_dict["unit"]),
        nullable=yaml_dict.get("nullable", True),
    ),
    "DateUnit": lambda yaml_dict: tm.DateUnit(yaml_dict["value"]),
    "TimeUnit": lambda yaml_dict: tm.TimeUnit(yaml_dict["value"]),
    "TimestampUnit": lambda yaml_dict: tm.TimestampUnit(yaml_dict["value"]),
    "Array": lambda yaml_dict: dt.Array(
        _type_from_yaml(yaml_dict["value_type"]),
        nullable=yaml_dict.get("nullable", True),
    ),
    "Map": lambda yaml_dict: dt.Map(
        _type_from_yaml(yaml_dict["key_type"]),
        _type_from_yaml(yaml_dict["value_type"]),
        nullable=yaml_dict.get("nullable", True),
    ),
}



import pytest

import xorq as xo
from xorq.caching import SourceStorage


@pytest.mark.parametrize(
    "name",
    (
        "diamonds",
        "hn-data-small.parquet",
    ),
)
def test_source_caching(name, pg):
    con = xo.connect()
    example = xo.examples.get_table_from_name(name, con)
    expr = example.cache(storage=SourceStorage(pg))
    assert not expr.ls.exists()
    actual = expr.execute()
    expected = example.execute()
    cached = pg.table(expr.ls.get_key()).execute()
    assert actual.equals(expected)
    assert actual.equals(cached)
    assert expr.ls.exists()
    pg.drop_table(expr.ls.get_key())
    assert not expr.ls.exists()



import pyarrow as pa

import xorq as xo
from xorq.flight import FlightServer
from xorq.flight.action import DropTableAction, ListTablesAction


def test_list_tables_kwargs():
    with FlightServer(
        verify_client=False,
        connection=xo.duckdb.connect,
    ) as main:
        # GIVEN
        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()
        main.con.register(data, table_name="users")
        expected = ["users"]

        # WHEN
        actual = main.client.do_action_one(ListTablesAction.name)

        # THEN
        assert actual == expected


def test_drop_table():
    with FlightServer(
        verify_client=False,
        connection=xo.duckdb.connect,
    ) as main:
        # GIVEN
        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()
        main.con.register(data, table_name="users")
        assert main.client.do_action_one(ListTablesAction.name)

        # WHEN
        main.client.do_action_one(DropTableAction.name, action_body={"name": "users"})

        # THEN
        assert not main.client.do_action_one(ListTablesAction.name)



import os
import pathlib
from typing import Any, Optional, Union

from xorq.vendor import ibis
from xorq.vendor.ibis.config import Config


class Cache(Config):
    """xorq cache configuration options

    Attributes
    ----------

    default_path : str

    """

    default_path: Union[str, pathlib.Path] = pathlib.Path(
        "~/.local/share/xorq"
    ).expanduser()
    key_prefix: str = "letsql_cache-"


class Interactive(Config):
    """Options controlling the interactive repr."""

    @property
    def max_rows(self) -> int:
        return ibis.options.repr.interactive.max_rows

    @max_rows.setter
    def max_rows(self, value: int):
        ibis.options.repr.interactive.max_rows = value

    @property
    def max_columns(self) -> Optional[int]:
        return ibis.options.repr.interactive.max_columns

    @max_columns.setter
    def max_columns(self, value: Optional[int]):
        ibis.options.repr.interactive.max_columns = value

    @property
    def max_length(self) -> int:
        return ibis.options.repr.interactive.max_length

    @max_length.setter
    def max_length(self, value: int):
        ibis.options.repr.interactive.max_length = value

    @property
    def max_string(self) -> int:
        return ibis.options.repr.interactive.max_string

    @max_string.setter
    def max_string(self, value: int):
        ibis.options.repr.interactive.max_string = value

    @property
    def max_depth(self) -> int:
        return ibis.options.repr.interactive.max_depth

    @max_depth.setter
    def max_depth(self, value: int):
        ibis.options.repr.interactive.max_depth = value

    @property
    def show_types(self) -> bool:
        return ibis.options.repr.interactive.show_types

    @show_types.setter
    def show_types(self, value: bool):
        ibis.options.repr.interactive.show_types = value


class Repr(Config):
    """Expression printing options.

    Attributes
    ----------
    interactive : Interactive
        Options controlling the interactive repr.
    """

    interactive: Interactive = Interactive()


class SQL(Config):
    """SQL-related options.

    Attributes
    ----------
    dialect : str
        Dialect to use for printing SQL when the backend cannot be determined.

    """

    dialect: str = "datafusion"


class Pins(Config):
    """SQL-related options.

    Attributes
    ----------
    dialect : str
        Dialect to use for printing SQL when the backend cannot be determined.

    """

    protocol: str = "gcs"
    path: str = "letsql-pins"
    storage_options: dict[str, Any] = dict(
        (
            ("cache_timeout", 0),
            ("token", "anon"),
        )
    )

    def get_board(self, **kwargs):
        import pins

        _kwargs = {
            **{
                "protocol": self.protocol,
                "path": self.path,
                "storage_options": self.storage_options,
            },
            **kwargs,
        }
        return pins.board(**_kwargs)

    def get_path(self, name, board=None):
        board = board or self.get_board()
        (path,) = board.pin_download(name)
        return path


class Profiles(Config):
    profile_dir: pathlib.Path = pathlib.Path("~/.config/xorq/profiles").expanduser()


class Options(Config):
    """xorq configuration options

    Attributes
    ----------
    cache : Cache
        Options controlling caching.
    backend : Optional[xorq.backends.let.Backend]
        The backend to use for execution.
    repr : Repr
        Options controlling expression printing.
    """

    cache: Cache = Cache()
    backend: Optional[Any] = None
    repr: Repr = Repr()
    sql: SQL = SQL()
    pins: Pins = Pins()
    profiles: Profiles = Profiles()
    debug: bool = bool(os.environ.get("XORQ_DEBUG", False))

    @property
    def interactive(self) -> bool:
        """Show the first few rows of computing an expression when in a repl."""
        return ibis.options.interactive

    @interactive.setter
    def interactive(self, value: bool):
        ibis.options.interactive = value


options = Options()


def _backend_init():
    if (backend := options.backend) is not None:
        return backend

    import xorq as xo

    options.backend = con = xo.connect()
    return con



import pyarrow as pa
import pytest

import xorq as xo


expected_tables = (
    "array_types",
    "astronauts",
    "awards_players",
    "awards_players_special_types",
    "batting",
    "diamonds",
    "functional_alltypes",
    "geo",
    "geography_columns",
    "geometry_columns",
    "json_t",
    "map",
    "spatial_ref_sys",
    "topk",
    "tzone",
)


@pytest.fixture(scope="session")
def pg():
    conn = xo.postgres.connect(
        host="localhost",
        port=5432,
        user="postgres",
        password="postgres",
        database="ibis_testing",
    )
    yield conn
    remove_unexpected_tables(conn)


@pytest.fixture(scope="session")
def dirty(pg):
    return pg


def remove_unexpected_tables(dirty):
    # drop tables
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_table(table, force=True)

    # drop view
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_view(table, force=True)

    if sorted(dirty.list_tables()) != sorted(expected_tables):
        raise ValueError


@pytest.fixture(scope="function")
def con(dirty):
    remove_unexpected_tables(dirty)
    yield dirty
    # cleanup
    remove_unexpected_tables(dirty)


@pytest.fixture(scope="session")
def dirty_ls_con():
    con = xo.connect()
    return con


@pytest.fixture(scope="function")
def ls_con(dirty_ls_con):
    # since we don't register, maybe just create a fresh con
    yield dirty_ls_con
    # drop tables
    for table_name in dirty_ls_con.list_tables():
        dirty_ls_con.drop_table(table_name, force=True)
    # drop view
    for table_name in dirty_ls_con.list_tables():
        dirty_ls_con.drop_view(table_name, force=True)


@pytest.fixture(scope="session")
def alltypes(dirty):
    return dirty.table("functional_alltypes")


@pytest.fixture(scope="session")
def batting(dirty):
    return dirty.table("batting")


@pytest.fixture(scope="session")
def awards_players(dirty):
    return dirty.table("awards_players")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def batting_df(batting):
    return batting.execute()


@pytest.fixture
def df():
    # create a RecordBatch and a new DataFrame from it
    batch = pa.RecordBatch.from_arrays(
        [
            pa.array([0, 1, 2, 3, 4, 5, 6]),
            pa.array([7, 4, 3, 8, 9, 1, 6]),
            pa.array(["A", "A", "A", "A", "B", "B", "B"]),
        ],
        names=["a", "b", "c"],
    )

    return batch.to_pandas()
